r"""
This module contains linear algebra solvers for SparseMatrixes
"""
import numpy as np
import scipy.sparse as scp
from scipy.sparse.linalg import spsolve, splu
from shenfun.optimization import optimizer
from shenfun.matrixbase import SparseMatrix, extract_bc_matrices, SpectralMatrix


class TDMA:
    """Tridiagonal matrix solver

    Parameters
    ----------
    mat : SparseMatrix
        Tridiagonal matrix with diagonals in offsets -2, 0, 2

    """
    # pylint: disable=too-few-public-methods

    def __init__(self, mat, neumann=False):
        assert isinstance(mat, SparseMatrix)
        self.mat = mat
        self.dd = np.zeros(0)
        self.neumann = neumann
        if isinstance(mat, SpectralMatrix):
            self.neumann = mat.testfunction[0].use_fixed_gauge

    def init(self):
        """Initialize and allocate solver"""
        N = self.mat.shape[0]
        self.symmetric = self.mat.issymmetric
        self.dd = self.mat[0]*np.ones(N)*self.mat.scale
        self.ud = self.mat[2]*np.ones(N-2)*self.mat.scale
        if self.neumann:
            self.dd[0] = 1
            self.ud[0] = 0
        self.ld = np.zeros(N-2) if self.symmetric else self.mat[-2]*np.ones(N-2)*self.mat.scale
        if self.symmetric:
            self.TDMA_SymLU(self.dd, self.ud, self.ld)
        else:
            self.TDMA_LU(self.ld, self.dd, self.ud)

    @staticmethod
    @optimizer
    def TDMA_LU(ld, d, ud):
        n = d.shape[0]
        for i in range(2, n):
            ld[i-2] = ld[i-2]/d[i-2]
            d[i] = d[i] - ld[i-2]*ud[i-2]

    @staticmethod
    @optimizer
    def TDMA_SymLU(d, ud, ld):
        n = d.shape[0]
        for i in range(2, n):
            ld[i-2] = ud[i-2]/d[i-2]
            d[i] = d[i] - ld[i-2]*ud[i-2]

    @staticmethod
    @optimizer
    def TDMA_SymSolve(d, a, l, x, axis=0):
        assert x.ndim == 1, "Use optimized version for multidimensional solve"
        n = d.shape[0]
        for i in range(2, n):
            x[i] -= l[i-2]*x[i-2]

        x[n-1] = x[n-1]/d[n-1]
        x[n-2] = x[n-2]/d[n-2]
        for i in range(n - 3, -1, -1):
            x[i] = (x[i] - a[i]*x[i+2])/d[i]

    def __call__(self, b, u=None, axis=0, **kw):
        """Solve matrix problem self u = b

        Parameters
        ----------
            b : array
                Array of right hand side on entry and solution on exit unless
                u is provided.
            u : array, optional
                Output array
            axis : int, optional
                   The axis over which to solve for if b and u are multidimensional

        Note
        ----
        If u is not provided, then b is overwritten with the solution and returned

        """

        if u is None:
            u = b
        else:
            assert u.shape == b.shape
            N = self.dd.shape[0]
            u[:] = b[:]

        if not self.dd.shape[0] == self.mat.shape[0]:
            self.init()

        if self.neumann:
            T = self.mat.testfunction[0]
            u[T.si[0]] = 0
            if isinstance(self, SpectralMatrix):
                u[T.si[0]] = self.mat.testfunction[0].mean

        self.TDMA_SymSolve(self.dd, self.ud, self.ld, u, axis=axis)

        if hasattr(u, 'set_boundary_dofs'):
            u.set_boundary_dofs()

        if self.neumann:
            u[T.si[0]] = 0
            if isinstance(self.mat, SpectralMatrix):
                u[T.si[0]] = self.mat.testfunction[0].mean

        return u

class PDMA:
    """Pentadiagonal matrix solver

    Parameters
    ----------
        mat : SparseMatrix
            Pentadiagonal matrix with diagonals in offsets
            -4, -2, 0, 2, 4
        neumann : bool, optional
            Whether matrix represents a Neumann problem, where
            the first index is known as the mean value and we
            solve for slice(1, N-3).
            If `mat` is a :class:`.SpectralMatrix`, then the
            `neumann` keyword is ignored and the information
            extracted from the matrix.

    """

    def __init__(self, mat, neumann=False):
        assert isinstance(mat, SparseMatrix)
        self.mat = mat
        self.N = 0
        self.d0 = np.zeros(0)
        self.d1 = None
        self.d2 = None
        self.A = None
        self.L = None
        self.neumann = neumann
        if isinstance(mat, SpectralMatrix):
            self.neumann = mat.testfunction[0].use_fixed_gauge

    def init(self):
        """Initialize and allocate solver"""
        B = self.mat
        shape = self.mat.shape[1]
        # Broadcast in case diagonal is simply a constant.
        self.d0 = np.broadcast_to(np.atleast_1d(B[0]), shape).copy()*B.scale
        self.d1 = np.broadcast_to(np.atleast_1d(B[2]), shape-2).copy()*B.scale
        self.d2 = np.broadcast_to(np.atleast_1d(B[4]), shape-4).copy()*B.scale
        if self.neumann:
            self.d0[0] = 1
            self.d1[0] = 0
            self.d2[0] = 0
        if B.issymmetric:
            self.PDMA_SymLU(self.d0, self.d1, self.d2)
        else:
            self.l1 = np.broadcast_to(np.atleast_1d(B[-2]), shape-2).copy()*B.scale
            self.l2 = np.broadcast_to(np.atleast_1d(B[-4]), shape-4).copy()*B.scale
            self.PDMA_LU(self.l2, self.l1, self.d0, self.d1, self.d2)

    @staticmethod
    @optimizer
    def PDMA_SymLU(d, e, f): # pragma: no cover
        """Symmetric LU decomposition"""
        n = d.shape[0]
        m = e.shape[0]
        k = n - m

        for i in range(n-2*k):
            lam = e[i]/d[i]
            d[i+k] -= lam*e[i]
            e[i+k] -= lam*f[i]
            e[i] = lam
            lam = f[i]/d[i]
            d[i+2*k] -= lam*f[i]
            f[i] = lam

        lam = e[n-4]/d[n-4]
        d[n-2] -= lam*e[n-4]
        e[n-4] = lam
        lam = e[n-3]/d[n-3]
        d[n-1] -= lam*e[n-3]
        e[n-3] = lam

    @staticmethod
    @optimizer
    def PDMA_LU(a, b, d, e, f): # pragma: no cover
        """LU decomposition"""
        n = d.shape[0]
        m = e.shape[0]
        k = n - m

        for i in range(n-2*k):
            lam = b[i]/d[i]
            d[i+k] -= lam*e[i]
            e[i+k] -= lam*f[i]
            b[i] = lam
            lam = a[i]/d[i]
            b[i+k] -= lam*e[i]
            d[i+2*k] -= lam*f[i]
            a[i] = lam

        i = n-4
        lam = b[i]/d[i]
        d[i+k] -= lam*e[i]
        b[i] = lam
        i = n-3
        lam = b[i]/d[i]
        d[i+k] -= lam*e[i]
        b[i] = lam

    @staticmethod
    @optimizer
    def PDMA_SymSolve(d, e, f, b, axis=0): # pragma: no cover
        """Symmetric solve (for testing only)"""
        n = d.shape[0]
        bc = b

        bc[2] -= e[0]*bc[0]
        bc[3] -= e[1]*bc[1]
        for k in range(4, n):
            bc[k] -= (e[k-2]*bc[k-2] + f[k-4]*bc[k-4])

        bc[n-1] /= d[n-1]
        bc[n-2] /= d[n-2]
        bc[n-3] /= d[n-3]
        bc[n-3] -= e[n-3]*bc[n-1]
        bc[n-4] /= d[n-4]
        bc[n-4] -= e[n-4]*bc[n-2]
        for k in range(n-5, -1, -1):
            bc[k] /= d[k]
            bc[k] -= (e[k]*bc[k+2] + f[k]*bc[k+4])
        b[:] = bc.astype(float)

    @staticmethod
    @optimizer
    def PDMA_Solve(a, b, d, e, f, h, axis=0): # pragma: no cover
        """Symmetric solve (for testing only)"""
        n = d.shape[0]
        bc = h

        bc[2] -= b[0]*bc[0]
        bc[3] -= b[1]*bc[1]
        for k in range(4, n):
            bc[k] -= (b[k-2]*bc[k-2] + a[k-4]*bc[k-4])

        bc[n-1] /= d[n-1]
        bc[n-2] /= d[n-2]
        bc[n-3] = (bc[n-3]-e[n-3]*bc[n-1])/d[n-3]
        bc[n-4] = (bc[n-4]-e[n-4]*bc[n-2])/d[n-4]
        for k in range(n-5, -1, -1):
            bc[k] = (bc[k]-e[k]*bc[k+2]-f[k]*bc[k+4])/d[k]

    def __call__(self, b, u=None, axis=0, **kw):
        """Solve matrix problem self u = b

        Parameters
        ----------
            b : array
                Array of right hand side on entry and solution on exit unless
                u is provided.
            u : array, optional
                Output array
            axis : int, optional
                   The axis over which to solve for if b and u are multidimensional

        Note
        ----
        If u is not provided, then b is overwritten with the solution and returned
        """

        if u is None:
            u = b
        else:
            assert u.shape == b.shape
            u[:] = b[:]

        if not self.d0.shape[0] == self.mat[0].shape[0]:
            self.init()

        if self.neumann:
            T = self.mat.testfunction[0]
            u[T.si[0]] = 0
            if isinstance(self, SpectralMatrix):
                u[T.si[0]] = T.mean

        if self.mat.issymmetric:
            self.PDMA_SymSolve(self.d0, self.d1, self.d2, u, axis)
        else:
            self.PDMA_Solve(self.l2, self.l1, self.d0, self.d1, self.d2, u, axis)

        if hasattr(u, 'set_boundary_dofs'):
            u.set_boundary_dofs()

        return u

class Solve:
    """Generic solver class for SparseMatrix

    Possibly with inhomogeneous boundary values

    Parameters
    ----------
        A : SparseMatrix
        test : BasisFunction

   """
    def __init__(self, A, test):
        assert isinstance(A, (SparseMatrix, list))
        self.s = test.slice()
        self.bc_mats = []
        if isinstance(A, list):
            bc_mats = extract_bc_matrices([A])
            self.A = np.sum(np.array(A, dtype=object))
            self.bc_mats = bc_mats
        else:
            self.A = A

        self.test = test
        assert self.A.shape[0] == self.A.shape[1]

    def __call__(self, b, u=None, axis=0, use_lu=False):
        """Solve matrix problem Au = b

        Parameters
        ----------
        b : array
            Array of right hand side on entry and solution on exit unless
            u is provided.
        u : array, optional
            Output array
        axis : int, optional
            The axis over which to solve for if b and u are multidimensional
        use_lu : bool, optional
            whether to use lu-decomposition stored previously as self._lu

        Note
        ----
        If u is not provided, then b is overwritten with the solution and returned

        """
        if u is None:
            u = b
        else:
            assert u.shape == b.shape

        # Boundary conditions
        if len(self.bc_mats) > 0:
            u.set_boundary_dofs()
            w0 = np.zeros_like(u)
            for bc_mat in self.bc_mats:
                b -= bc_mat.matvec(u, w0)

        # Move axis to first
        if axis > 0:
            u = np.moveaxis(u, axis, 0)
            if u is not b:
                b = np.moveaxis(b, axis, 0)

        s = self.s
        assert self.A.shape[0] == b[s].shape[0]

        if self.test.use_fixed_gauge:
            A = self.A.diags('lil')
            _, zerorow = A[0].nonzero()
            A[(0, zerorow)] = 0
            A[0, 0] = 1
            b[0] = self.test.mean
            A = A.tocsc()
        else:
            A = self.A.diags('csc')

        if b.ndim == 1:
            if use_lu:
                if b.dtype.char in 'fdg' or self._lu.U.dtype.char in 'FDG':
                    u[s] = self._lu.solve(b[s])
                else: # complex b and real matrix
                    u.real[s] = self._lu.solve(b[s].real)
                    u.imag[s] = self._lu.solve(b[s].imag)

            else:
                u[s] = spsolve(A, b[s])

        else:
            N = b[s].shape[0]
            P = np.prod(b[s].shape[1:])
            br = b[s].reshape((N, P))
            if use_lu:
                if b.dtype.char in 'fdg' or self._lu.U.dtype.char in 'FDG':
                    u[s] = self._lu.solve(br).reshape(u[s].shape)
                else: # complex b and real matrix
                    u.real[s] = self._lu.solve(br.real).reshape(u[s].shape)
                    u.imag[s] = self._lu.solve(br.imag).reshape(u[s].shape)

            else:
                u[s] = spsolve(A, br).reshape(u[s].shape)

        if axis > 0:
            u = np.moveaxis(u, 0, axis)
            if u is not b:
                b = np.moveaxis(b, 0, axis)
        return u


class SolverGeneric2ND:
    """Generic solver for problems consisting of tensorproduct matrices
    containing two non-diagonal submatrices.

    Parameters
    ----------
    mats : sequence
        sequence of instances of :class:`.TPMatrix`

    Note
    ----
    In addition to two non-diagonal matrices, the solver can also handle one
    additional diagonal matrix (one Fourier matrix).
    """

    def __init__(self, tpmats):
        self.tpmats = tpmats
        self.T = T = tpmats[0].space
        self.M = None

    def matvec(self, u, c):
        c.fill(0)
        if u.ndim == 2:
            s0 = tuple(base.slice() for base in self.T)
            c[s0] = self.M.dot(u[s0].flatten()).reshape(self.T.dims())
        return c

    def get_diagonal_axis(self):
        naxes = self.T.get_nondiagonal_axes()
        diagonal_axis = np.setxor1d([0, 1, 2], naxes)
        assert len(diagonal_axis) == 1
        return diagonal_axis[0]

    def diags(self, i):
        """Return matrix for given index `i` in diagonal direction"""
        if self.T.dimensions == 2:
            # In 2D there's just 1 matrix, store and reuse
            if self.M is not None:
                return self.M
            m = self.tpmats[0]
            A0 = m.mats[0].diags('csc')
            A1 = m.mats[1].diags('csc')
            M0 = scp.kron(A0, A1, 'csc')
            M0 *= np.atleast_1d(m.scale).item()
            for m in self.tpmats[1:]:
                A0 = m.mats[0].diags('csc')
                A1 = m.mats[1].diags('csc')
                M1 = scp.kron(A0, A1, 'csc')
                M1 *= np.atleast_1d(m.scale).item()
                M0 = M0 + M1
            # Check if we need to fix gauge. This is required if we are solving
            # a pure Neumann Poisson problem.
            z0 = M0[0].nonzero()
            if z0[1][0] > 0 :
                M0 = M0.tolil()
                zerorow = M0[0].nonzero()[1]
                M0[(0, zerorow)] = 0
                M0[0, 0] = 1
                M0 = M0.tocsc()
            self.M = M0
            return self.M

        else:
            # 1 matrix per Fourier coefficient
            naxes = self.T.get_nondiagonal_axes()
            m = self.tpmats[0]
            diagonal_axis = self.get_diagonal_axis()
            sc = [0, 0, 0]
            sc[diagonal_axis] = i if m.scale.shape[diagonal_axis] > 1 else 0
            A0 = m.mats[naxes[0]].diags('csc')
            A1 = m.mats[naxes[1]].diags('csc')
            M0 = scp.kron(A0, A1, 'csc')
            M0 *= m.scale[tuple(sc)]
            for m in self.tpmats[1:]:
                A0 = m.mats[naxes[0]].diags('csc')
                A1 = m.mats[naxes[1]].diags('csc')
                M1 = scp.kron(A0, A1, 'csc')
                sc[diagonal_axis] = i if m.scale.shape[diagonal_axis] > 1 else 0
                M1 *= m.scale[tuple(sc)]
                M0 = M0 + M1
            z0 = M0[0].nonzero()
            if z0[1][0] > 0 :
                M0 = M0.tolil()
                zerorow = M0[0].nonzero()[1]
                M0[(0, zerorow)] = 0
                M0[0, 0] = 1
                M0 = M0.tocsc()
            return M0

    def __call__(self, b, u=None):
        if u is None:
            u = b
        else:
            assert u.shape == b.shape
        if u.ndim == 2:
            s0 = self.T.slice()
            M = self.diags(0)
            u[s0] = scp.linalg.spsolve(M, b[s0].flatten()).reshape(self.T.dims())

        elif u.ndim == 3:
            naxes = self.T.get_nondiagonal_axes()
            diagonal_axis = self.get_diagonal_axis()
            s0 = list(self.T.slice())
            for i in range(self.T.shape(True)[diagonal_axis]):
                M0 = self.diags(i)
                s0[diagonal_axis] = i
                shape = np.take(self.T.dims(), naxes)
                u[tuple(s0)] = scp.linalg.spsolve(M0, b[tuple(s0)].flatten()).reshape(shape)
        return u

class Solver2D:
    """Generic solver for tensorproductspaces in 2D

    Parameters
    ----------
    mats : sequence
        sequence of instances of :class:`.TPMatrix`

    Note
    ----
    If there are boundary matrices in the list of mats, then
    these matrices are used to modify the right hand side before
    solving. If this is not the desired behaviour, then use
    :func:`.extract_bc_matrices` on mats before using this class.

    """

    def __init__(self, tpmats, fixed_gauge=None):
        bc_mats = extract_bc_matrices([tpmats])
        self.tpmats = tpmats
        self.bc_mats = bc_mats
        self.fixed_gauge = fixed_gauge
        self._lu = None
        m = tpmats[0]
        self.T = T = m.space
        ndim = T.dimensions
        assert ndim == 2
        assert np.atleast_1d(m.scale).size == 1, "Use level = 2 with :func:`.inner`"

        A0 = m.mats[0].diags('csc')
        A1 = m.mats[1].diags('csc')
        M0 = scp.kron(A0, A1, 'csc')
        M0 *= np.atleast_1d(m.scale).item()
        for m in tpmats[1:]:
            A0 = m.mats[0].diags('csc')
            A1 = m.mats[1].diags('csc')
            M1 = scp.kron(A0, A1, 'csc')
            assert np.atleast_1d(m.scale).size == 1, "Use level = 2 with :func:`.inner`"
            M1 *= np.atleast_1d(m.scale).item()
            M0 = M0 + M1

        # Check if we need to fix gauge. This is required if we are solving
        # a pure Neumann Poisson problem.
        if fixed_gauge is not None:
            # Ident first row
            z0 = M0[0].nonzero()
            if len(z0[0]) == 0 or z0[1][0] > 0:
                M0 = M0.tolil()
                zerorow = M0[0].nonzero()[1]
                M0[(0, zerorow)] = 0
                M0[0, 0] = 1
                M0 = M0.tocsc()

        self.M = M0

    def matvec(self, u, c):
        c.fill(0)
        s0 = tuple(base.slice() for base in self.T)
        c[s0] = self.M.dot(u[s0].flatten()).reshape(self.T.dims())
        return c

    def __call__(self, b, u=None):

        if u is None:
            u = b
        else:
            assert u.shape == b.shape

        if len(self.bc_mats) > 0:
            u.set_boundary_dofs()
            w0 = np.zeros_like(u)
            for bc_mat in self.bc_mats:
                b -= bc_mat.matvec(u, w0)

        s0 = tuple(base.slice() for base in self.T)
        assert b.dtype.char == u.dtype.char
        if self._lu is None:
            self._lu = scp.linalg.splu(self.M)
        if b.dtype.char in 'fdg':
            u[s0] = self._lu.solve(b[s0].flatten()).reshape(self.T.dims())
        else:
            if self.M.dtype.char in 'FDG':
                u[s0] = self._lu.solve(b[s0].flatten()).reshape(self.T.dims())

            else:
                u.imag[s0] = self._lu.solve(b.imag[s0].flatten()).reshape(self.T.dims())
                u.real[s0] = self._lu.solve(b.real[s0].flatten()).reshape(self.T.dims())

        if self.fixed_gauge:
            u[0, 0] = self.fixed_gauge

        return u

class Solver3D:
    """Generic solver for tensorproductspaces in 3D

    Parameters
    ----------
    mats : sequence
        sequence of instances of :class:`.TPMatrix`

    Note
    ----
    If there are boundary matrices in the list of mats, then
    these matrices are used to modify the right hand side before
    solving. If this is not the desired behaviour, then use
    :func:`.extract_bc_matrices` on mats before using this class.

    """

    def __init__(self, tpmats, fixed_gauge=None):
        bc_mats = extract_bc_matrices([tpmats])
        self.tpmats = tpmats
        self.bc_mats = bc_mats
        self._lu = None
        m = tpmats[0]
        self.T = T = m.space # test space
        ndim = T.dimensions
        assert ndim == 3
        assert np.atleast_1d(m.scale).size == 1, "Use level = 2 with :func:`.inner`"

        A0 = m.mats[0].diags('csc')
        A1 = m.mats[1].diags('csc')
        A2 = m.mats[2].diags('csc')
        M0 = scp.kron(scp.kron(A0, A1, 'csc'), A2, 'csc')
        M0 *= np.atleast_1d(m.scale).item()
        for m in tpmats[1:]:
            A0 = m.mats[0].diags('csc')
            A1 = m.mats[1].diags('csc')
            A2 = m.mats[2].diags('csc')
            M1 = scp.kron(scp.kron(A0, A1, 'csc'), A2, 'csc')
            assert np.atleast_1d(m.scale).size == 1, "Use level = 2 with :func:`.inner`"
            M1 *= np.atleast_1d(m.scale).item()
            M0 = M0 + M1

        # Check if we need to fix gauge. This is required if we are solving
        # a pure Neumann Poisson problem.
        if fixed_gauge:
            # Ident first row
            z0 = M0[0].nonzero()
            if len(z0[0]) == 0 or z0[1][0] > 0:
                M0 = M0.tolil()
                zerorow = M0[0].nonzero()[1]
                M0[(0, zerorow)] = 0
                M0[0, 0] = 1
                M0 = M0.tocsc()
        self.M = M0

    def matvec(self, u, c):
        c.fill(0)
        s0 = tuple(base.slice() for base in self.T)
        c[s0] = self.M.dot(u[s0].flatten()).reshape(self.T.dims())
        return c

    def __call__(self, b, u=None):

        if u is None:
            u = b
        else:
            assert u.shape == b.shape

        if len(self.bc_mats) > 0:
            u.set_boundary_dofs()
            w0 = np.zeros_like(u)
            for bc_mat in self.bc_mats:
                b -= bc_mat.matvec(u, w0)

        s0 = tuple(base.slice() for base in self.T)
        assert b.dtype.char == u.dtype.char
        if self._lu is None:
            self._lu = scp.linalg.splu(self.M)
        if b.dtype.char in 'fdg':
            u[s0] = self._lu.solve(b[s0].flatten()).reshape(self.T.dims())
        else:
            if self.M.dtype.char in 'FDG':
                u[s0] = self._lu.solve(b[s0].flatten()).reshape(self.T.dims())

            else:
                u.imag[s0] = self._lu.solve(b.imag[s0].flatten()).reshape(self.T.dims())
                u.real[s0] = self._lu.solve(b.real[s0].flatten()).reshape(self.T.dims())

        if self.fixed_gauge:
            u[0, 0, 0] = self.fixed_gauge

        return u

class TDMA_O:
    """Tridiagonal matrix solver

    Parameters
    ----------
    mat : SparseMatrix
        Symmetric tridiagonal matrix with diagonals in offsets -1, 0, 1

    """
    # pylint: disable=too-few-public-methods

    def __init__(self, mat):
        assert isinstance(mat, SparseMatrix)
        self.mat = mat
        self.N = 0
        self.dd = np.zeros(0)
        self.ud = None
        self.L = None

    def init(self):
        """Initialize and allocate solver"""
        M = self.mat.shape[0]
        B = self.mat
        self.dd = B[0]*np.ones(M)
        self.ud = B[1]*np.ones(M-1)
        self.L = np.zeros(M-1)
        self.TDMA_O_SymLU(self.dd, self.ud, self.L)

    @staticmethod
    @optimizer
    def TDMA_O_SymLU(d, ud, ld):
        n = d.shape[0]
        for i in range(1, n):
            ld[i-1] = ud[i-1]/d[i-1]
            d[i] = d[i] - ld[i-1]*ud[i-1]

    @staticmethod
    @optimizer
    def TDMA_O_SymSolve(d, a, l, x, axis=0):
        assert x.ndim == 1, "Use optimized version for multidimensional solve"
        n = d.shape[0]
        for i in range(1, n):
            x[i] -= l[i-1]*x[i-1]

        x[n-1] = x[n-1]/d[n-1]
        for i in range(n-2, -1, -1):
            x[i] = (x[i] - a[i]*x[i+1])/d[i]

    def __call__(self, b, u=None, axis=0):
        """Solve matrix problem self u = b

        Parameters
        ----------
            b : array
                Array of right hand side on entry and solution on exit unless
                u is provided.
            u : array, optional
                Output array
            axis : int, optional
                   The axis over which to solve for if b and u are multidimensional

        Note
        ----
        If u is not provided, then b is overwritten with the solution and returned

        """

        if u is None:
            u = b
        else:
            assert u.shape == b.shape
            u[:] = b[:]

        if not self.dd.shape[0] == self.mat.shape[0]:
            self.init()

        self.TDMA_O_SymSolve(self.dd, self.ud, self.L, u, axis=axis)

        u /= self.mat.scale
        return u


class SolverGeneric1ND:
    """Generic solver for tensorproduct matrices consisting of
    non-diagonal matrices along only one axis.

    Parameters
    ----------
    mats : sequence
        sequence of instances of :class:`.TPMatrix`

    Note
    ----
    In addition to the one non-diagonal direction, the solver can also handle
    up to two diagonal (Fourier) directions.
    Also note that if there are boundary matrices in the list of mats, then
    these matrices are used to modify the right hand side before
    solving. If this is not the desired behaviour, then use
    :func:`.extract_bc_matrices` on mats before using this class.

    """

    def __init__(self, mats):
        assert isinstance(mats, list)
        naxes = set()
        for tpmat in mats:
            if not tpmat._issimplified:
                tpmat.simplify_diagonal_matrices()
            naxes.update(tpmat.naxes)
        assert len(naxes) == 1
        self.naxes = naxes.pop()
        bc_mats = extract_bc_matrices([mats])
        self.mats = mats
        self.bc_mats = bc_mats
        # For time-dependent solver, store all generated matrices and reuse
        # This takes a lot of memory, so for now it's only implemented for 2D
        self.MM = None

    @staticmethod
    def apply_constraint(A, b, i, constraint):
        assert isinstance(constraint, tuple)
        assert len(constraint) == 4

        if constraint is None:
            return A, b

        if not i == constraint[0]:
            return A, b

        row = constraint[1]
        col = constraint[2]
        val = constraint[3]
        b[row] = val
        r = A.getrow(row).nonzero()
        A[(row, r[1])] = 0
        A[row, col] = 1
        return A, b

    def matvec(self, u, c):
        c.fill(0)
        w0 = np.zeros_like(u)
        for mat in self.mats:
            c += mat.matvec(u, w0)

        if len(self.bc_mats) > 0:
            u.set_boundary_dofs()
            for bc_mat in self.bc_mats:
                c += bc_mat.matvec(u, w0)
        return c

    def __call__(self, b, u=None, constraints=()):
        """Solve problem with one non-diagonal direction

        Parameters
        ----------
        b : array, right hand side
        u : array, solution
        constraints : tuple of 4-tuples
            Each 4-tuple is a constraint, with each item representing
              - 0 : The diagonal index, or indices for 3D
              - 1 : row
              - 2 : column
              - 3 : value
            Matrix row is zeroed and then indented by setting A[row, col] = 0

        """
        if u is None:
            u = b
        else:
            assert u.shape == b.shape
        m = self.mats[0]

        if len(self.bc_mats) > 0:
            u.set_boundary_dofs()
            w0 = np.zeros_like(u)
            for bc_mat in self.bc_mats:
                b -= bc_mat.matvec(u, w0)

        if u.ndim == 2:

            if self.naxes == 0:
                # non-diagonal in axis=0

                if self.MM is None:
                    self.MM = []
                    for i in range(b.shape[1]):
                        MM = None
                        for mat in self.mats:
                            sc = mat.scale[0, i] if mat.scale.shape[1] > 1 else mat.scale[0, 0]
                            if MM:
                                MM += mat.mats[0]*sc
                            else:
                                MM = mat.mats[0]*sc
                        sl = m.space.bases[0].slice()
                        #u[sl, i] = MM.solve(b[sl, i], u[sl, i])
                        Mc = MM.diags('csc')
                        for constraint in constraints:
                            Mc, b = self.apply_constraint(Mc, b, i, constraint)
                        try:
                            MM._lu = splu(Mc)
                            u[sl, i] = MM.solve(b[sl, i], u[sl, i], use_lu=True)
                        except:
                            print('Singular matrix for j=', i)
                            u[sl, i] = 0
                        self.MM.append(MM)

                else:
                    for i in range(b.shape[1]):
                        sl = m.space.bases[0].slice()
                        u[sl, i] = self.MM[i].solve(b[sl, i], u[sl, i], use_lu=True)

            else:
                if self.MM is None:
                    # non-diagonal in axis=1
                    self.MM = []
                    for i in range(b.shape[0]):
                        MM = None
                        for mat in self.mats:
                            sc = mat.scale[i, 0] if mat.scale.shape[0] > 1 else mat.scale[0, 0]
                            if MM:
                                MM += mat.mats[1]*sc
                            else:
                                MM = mat.mats[1]*sc
                        sl = m.space.bases[1].slice()
                        Mc = MM.diags('csc')
                        for constraint in constraints:
                            Mc, b = self.apply_constraint(Mc, b, i, constraint)
                        MM._lu = splu(Mc)
                        MM.solve(b[i, sl], u[i, sl], use_lu=True)
                        self.MM.append(MM)

                else:
                    for i in range(b.shape[0]):
                        sl = m.space.bases[1].slice()
                        u[i, sl] = self.MM[i].solve(b[i, sl], u[i, sl], use_lu=True)

        elif u.ndim == 3:
            if self.naxes == 0:
                # non-diagonal in axis=0
                for i in range(b.shape[1]):
                    for j in range(b.shape[2]):
                        MM = None
                        for mat in self.mats:
                            sc = np.broadcast_to(mat.scale, u.shape)[0, i, j]
                            if MM:
                                MM += mat.mats[0]*sc
                            else:
                                MM = mat.mats[0]*sc
                        Mc = MM.diags('csc')
                        for constraint in constraints:
                            Mc, b = self.apply_constraint(Mc, b, (i, j), constraint)
                        MM._lu = splu(Mc)
                        sl = mat.space.bases[0].slice()
                        u[sl, i, j] = MM.solve(b[sl, i, j], u[sl, i, j], use_lu=True)

            elif self.naxes == 1:
                # non-diagonal in axis=1
                for i in range(b.shape[0]):
                    for j in range(b.shape[2]):
                        MM = None
                        for mat in self.mats:
                            sc = np.broadcast_to(mat.scale, u.shape)[i, 0, j]
                            if MM:
                                MM += mat.mats[1]*sc
                            else:
                                MM = mat.mats[1]*sc
                        Mc = MM.diags('csc')
                        for constraint in constraints:
                            Mc, b = self.apply_constraint(Mc, b, (i, j), constraint)
                        MM._lu = splu(Mc)
                        sl = mat.space.bases[1].slice()
                        u[i, sl, j] = MM.solve(b[i, sl, j], u[i, sl, j], use_lu=True)

            elif self.naxes == 2:
                # non-diagonal in axis=2
                for i in range(b.shape[0]):
                    for j in range(b.shape[1]):
                        MM = None
                        for mat in self.mats:
                            sc = np.broadcast_to(mat.scale, u.shape)[i, j, 0]
                            if MM:
                                MM += mat.mats[2]*sc
                            else:
                                MM = mat.mats[2]*sc
                        Mc = MM.diags('csc')
                        for constraint in constraints:
                            Mc, b = self.apply_constraint(Mc, b, (i, j), constraint)
                        MM._lu = splu(Mc)
                        sl = mat.space.bases[2].slice()
                        u[i, j, sl] = MM.solve(b[i, j, sl], u[i, j, sl], use_lu=True)

        return u



"""
Module for handling Fourier diagonal matrices
"""
from __future__ import division

#__all__ = ['mat']

import functools
import numpy as np
import sympy as sp
from shenfun.matrixbase import SpectralMatrix
from . import bases

R2C = bases.R2C
C2C = bases.C2C

xp = sp.Symbol('x', real=True, positive=True)


class Acos2mat(SpectralMatrix):
    r"""Matrix for inner product

    .. math::

        A_{kj}=(\exp(i k x), \exp(i l x)'')_w

    where weight w = \cos^2 x

    """
    def __init__(self, test, trial, measure=1):
        k = test[0].wavenumbers(bcast=False, scaled=False, eliminate_highest_freq=False)
        N = test[0].N
        d = {0: -0.5*k**2,
             2: -0.25*k[2:]**2,
             -2: -0.25*k[:-2]**2,
             N-2: -0.25*k[-2:]**2,
             -(N-2): -0.25*k[:2]**2}
        SpectralMatrix.__init__(self, d, test, trial, measure=measure)


class Acosmat(SpectralMatrix):
    r"""Matrix for inner product

    .. math::

        A_{kj}=(\exp(i k x), \exp(i l x)'')_w

    where weight w = \cos x

    """
    def __init__(self, test, trial, measure=1):
        k = test[0].wavenumbers(bcast=False, scaled=False, eliminate_highest_freq=False)
        N = test[0].N
        d = {1: -0.5*k[1:]**2,
             -1: -0.5*k[:-1]**2,
             N-1: -0.5*k[-1]**2}
        SpectralMatrix.__init__(self, d, test, trial, measure=measure)


class Csinmat(SpectralMatrix):
    r"""Matrix for inner product

    .. math::

        A_{kj}=(\exp(i k x), \exp(i l x)')_w

    where weight w = \sin x

    """
    def __init__(self, test, trial, measure=1):
        k = test[0].wavenumbers(bcast=False, scaled=False, eliminate_highest_freq=False)
        N = test[0].N
        d = {1: -0.5*k[1:],
             -1: 0.5*k[:-1],
             N-1: -0.5}
        SpectralMatrix.__init__(self, d, test, trial, measure=measure)


class Csincosmat(SpectralMatrix):
    r"""Matrix for inner product

    .. math::

        A_{kj}=(\exp(i k x), \exp(i l x)')_w

    where weight w = \sin x

    """
    def __init__(self, test, trial, measure=1):
        k = test[0].wavenumbers(bcast=False, scaled=False, eliminate_highest_freq=False)
        N = test[0].N
        d = {2: -0.25*k[2:],
             -2: 0.25*k[:-2],
             N-2: 0.25*k[-2:],
             -(N-2): -0.25*k[:2]}
        SpectralMatrix.__init__(self, d, test, trial, measure=measure)


class Bcos2mat(SpectralMatrix):
    r"""Matrix for inner product

    .. math::

        A_{kj}=(\exp(i k x), \exp(i l x))_w

    where weight w = \cos^2 x

    """
    def __init__(self, test, trial, measure=1):
        N = test[0].N
        d = {0: 0.5,
             2: 0.25,
             -2: 0.25,
             N-2: 0.25,
             -(N-2): 0.25}
        SpectralMatrix.__init__(self, d, test, trial, measure=measure)


class Bcosmat(SpectralMatrix):
    r"""Matrix for inner product

    .. math::

        A_{kj}=(\exp(i k x), \exp(i l x))_w

    where weight w = \cos x

    """
    def __init__(self, test, trial, measure=1):
        k = test[0].wavenumbers(bcast=False, scaled=False, eliminate_highest_freq=False)
        N = test[0].N
        d = {1: 0.5,
             -1: 0.5,
             N-1: 0.5,
             -(N-1): 0.5}
        SpectralMatrix.__init__(self, d, test, trial, measure=measure)


class _Fouriermatrix(SpectralMatrix):
    def __init__(self, test, trial, measure=1):
        N = test[0].N
        d = {}
        if measure == 1:
            k = test[0].wavenumbers(N, scaled=False)
            if isinstance(test[1], (int, np.integer)):
                k_test, k_trial = test[1], trial[1]
            elif isinstance(test[1], np.ndarray):
                assert len(test[1]) == 1
                k_test = test[1][(0,)*np.ndim(test[1])]
                k_trial = trial[1][(0,)*np.ndim(trial[1])]
            else:
                raise RuntimeError

            if abs(k_trial) + abs(k_test) > 0:
                if N % 2 == 0 and (k_trial + k_test) % 2 == 1:
                    pass
                    #k[N//2] = 0
                val = (1j*k)**(k_trial)*(-1j*k)**k_test
                if (k_trial + k_test) % 2 == 0:
                    val = val.real
                d = {0: val}
            else:
                d = {0: 1.0}
        SpectralMatrix.__init__(self, d, test, trial, measure=measure)

    def solve(self, b, u=None, axis=0):
        if self.measure == 1:
            N = self.shape[0]

            if u is None:
                u = b
            else:
                assert u.shape == b.shape

            with np.errstate(divide='ignore'):
                d = 1./self[0]
            if isinstance(d, np.ndarray):
                if np.isinf(d[0]):
                    d[0] = 0
                if np.isinf(d[N//2]):
                    d[N//2] = 0
                sl = [np.newaxis]*u.ndim
                sl[axis] = slice(None)
                u[:] = b*d[tuple(sl)]
            else:
                u[:] = b*d

            u /= self.scale
            return u

        return SpectralMatrix.solve(self, b, u=u, axis=axis)


class _FourierMatDict(dict):
    """Dictionary of inner product matrices.

    Matrices that are missing keys are generated. All Fourier matrices are
    diagonal.

    """

    def __missing__(self, key):
        measure = 1 if len(key) == 2 else key[3]
        c = functools.partial(_Fouriermatrix, measure=measure)
        self[key] = c
        return c

    def __getitem__(self, key):
        matrix = dict.__getitem__(self, key)
        return matrix


mat = _FourierMatDict({
    ((C2C, 0), (C2C, 2), (0, 2*np.pi), sp.cos(xp)**2): functools.partial(Acos2mat, measure=sp.cos(xp)**2),
    ((C2C, 0), (C2C, 2), (0, 2*np.pi), sp.cos(xp)): functools.partial(Acosmat, measure=sp.cos(xp)),
    ((C2C, 0), (C2C, 1), (0, 2*np.pi), sp.sin(xp)): functools.partial(Csinmat, measure=sp.sin(xp)),
    ((C2C, 0), (C2C, 1), (0, 2*np.pi), sp.sin(2*xp)/2): functools.partial(Csincosmat, measure=sp.sin(2*xp)/2),
    ((C2C, 0), (C2C, 1), (0, 2*np.pi), sp.sin(xp)*sp.cos(xp)): functools.partial(Csincosmat, measure=sp.sin(xp)*sp.cos(xp)),
    ((C2C, 0), (C2C, 0), (0, 2*np.pi), sp.cos(xp)**2): functools.partial(Bcos2mat, measure=sp.cos(xp)**2),
    ((C2C, 0), (C2C, 0), (0, 2*np.pi), sp.cos(xp)): functools.partial(Bcosmat, measure=sp.cos(xp)),
})

#mat = _FourierMatDict({})



#cython: boundscheck=False, wraparound=False, language_level=3

import numpy as np
cimport cython
cimport numpy as np
from libcpp.vector cimport vector
from libc.math cimport M_PI, M_PI_2
from cython.parallel import prange

ctypedef np.complex128_t complex_t
ctypedef np.float64_t real_t
ctypedef np.int64_t int_t

ctypedef fused T:
    real_t
    complex_t

def imult(T[:, :, ::1] array, real_t scale):
    cdef int i, j, k

    #for i in prange(array.shape[0], nogil=True, num_threads=1):
    for i in range(array.shape[0]):
        for j in range(array.shape[1]):
            for k in range(array.shape[2]):
                array[i, j, k] *= scale
    return array

def CDN_matvec1D_ptr(T[::1] v,
                     T[::1] b,
                     real_t[::1] ld,
                     real_t[::1] ud):
    cdef:
        T* v_ptr = &v[0]
        T* b_ptr = &b[0]
        int N = v.shape[0]-2
    CDN_matvec_ptr(v_ptr, b_ptr, &ld[0], &ud[0], N, 1)

cdef void CDN_matvec_ptr(T* v,
                         T* b,
                         real_t* ld,
                         real_t* ud,
                         int N,
                         int st):
    cdef:
        int i

    b[0] = ud[0]*v[st]
    b[(N-1)*st] = ld[N-2]*v[(N-2)*st]
    for i in xrange(1, N-1):
        b[i*st] = ud[i]*v[(i+1)*st] + ld[i-1]*v[(i-1)*st]

def CDN_matvec2D_ptr(T[:, ::1] v,
                     T[:, ::1] b,
                     real_t[::1] ld,
                     real_t[::1] ud,
                     int axis):
    cdef:
        int i, j, strides
        int N = v.shape[axis]-2
        T* v_ptr
        T* b_ptr

    strides = v.strides[axis]/v.itemsize
    if axis == 0:
        for j in range(v.shape[1]):
            v_ptr = &v[0, j]
            b_ptr = &b[0, j]
            CDN_matvec_ptr(v_ptr, b_ptr, &ld[0], &ud[0], N, strides)
    elif axis == 1:
        for i in range(v.shape[0]):
            v_ptr = &v[i, 0]
            b_ptr = &b[i, 0]
            CDN_matvec_ptr(v_ptr, b_ptr, &ld[0], &ud[0], N, strides)

def CDN_matvec3D_ptr(T[:, :, ::1] v,
                     T[:, :, ::1] b,
                     real_t[::1] ld,
                     real_t[::1] ud,
                     int axis):
    cdef:
        int i, j, k, strides
        int N = v.shape[axis]-2
        T* v_ptr
        T* b_ptr

    strides = v.strides[axis]/v.itemsize
    if axis == 0:
        for j in range(v.shape[1]):
            for k in range(v.shape[2]):
                v_ptr = &v[0, j, k]
                b_ptr = &b[0, j, k]
                CDN_matvec_ptr(v_ptr, b_ptr, &ld[0], &ud[0], N, strides)
    elif axis == 1:
        for i in range(v.shape[0]):
            for k in range(v.shape[2]):
                v_ptr = &v[i, 0, k]
                b_ptr = &b[i, 0, k]
                CDN_matvec_ptr(v_ptr, b_ptr, &ld[0], &ud[0], N, strides)
    elif axis == 2:
        for i in range(v.shape[0]):
            for j in range(v.shape[1]):
                v_ptr = &v[i, j, 0]
                b_ptr = &b[i, j, 0]
                CDN_matvec_ptr(v_ptr, b_ptr, &ld[0], &ud[0], N, strides)

cdef void BDN_matvec_ptr(T* v,
                         T* b,
                         real_t* ld,
                         real_t* dd,
                         real_t ud,
                         int N,
                         int st):
    cdef:
        int i, j, k

    b[0] = ud*v[2*st] + dd[0]*v[0]
    b[st] = ud*v[3*st] + dd[1]*v[st]
    b[(N-2)*st] = ld[N-4]*v[(N-4)*st] + dd[N-2]*v[(N-2)*st]
    b[(N-1)*st] = ld[N-3]*v[(N-3)*st] + dd[N-1]*v[(N-1)*st]
    for i in xrange(2, N-2):
        b[i*st] = ud*v[(i+2)*st] + dd[i]*v[i*st] + ld[i-2]*v[(i-2)*st]

def BDN_matvec1D_ptr(T[::1] v,
                     T[::1] b,
                     real_t[::1] ld,
                     real_t[::1] dd,
                     real_t ud):
    cdef:
        T* v_ptr = &v[0]
        T* b_ptr = &b[0]
        int N = v.shape[0]-2
    BDN_matvec_ptr(v_ptr, b_ptr, &ld[0], &dd[0], ud, N, 1)

def BDN_matvec3D_ptr(T[:, :, ::1] v,
                     T[:, :, ::1] b,
                     real_t[::1] ld,
                     real_t[::1] dd,
                     real_t ud,
                     int axis):
    cdef:
        int i, j, k, strides
        int N = v.shape[axis]-2
        T* v_ptr
        T* b_ptr

    strides = v.strides[axis]/v.itemsize
    if axis == 0:
        for j in range(v.shape[1]):
            for k in range(v.shape[2]):
                v_ptr = &v[0, j, k]
                b_ptr = &b[0, j, k]
                BDN_matvec_ptr(v_ptr, b_ptr, &ld[0], &dd[0], ud, N, strides)
    elif axis == 1:
        for i in range(v.shape[0]):
            for k in range(v.shape[2]):
                v_ptr = &v[i, 0, k]
                b_ptr = &b[i, 0, k]
                BDN_matvec_ptr(v_ptr, b_ptr, &ld[0], &dd[0], ud, N, strides)
    elif axis == 2:
        for i in range(v.shape[0]):
            for j in range(v.shape[1]):
                v_ptr = &v[i, j, 0]
                b_ptr = &b[i, j, 0]
                BDN_matvec_ptr(v_ptr, b_ptr, &ld[0], &dd[0], ud, N, strides)

def BDN_matvec2D_ptr(T[:, ::1] v,
                     T[:, ::1] b,
                     real_t[::1] ld,
                     real_t[::1] dd,
                     real_t ud,
                     int axis):
    cdef:
        int i, j, strides
        int N = v.shape[axis]-2
        T* v_ptr
        T* b_ptr

    strides = v.strides[axis]/v.itemsize
    if axis == 0:
        for j in range(v.shape[1]):
            v_ptr = &v[0, j]
            b_ptr = &b[0, j]
            BDN_matvec_ptr(v_ptr, b_ptr, &ld[0], &dd[0], ud, N, strides)
    elif axis == 1:
        for i in range(v.shape[0]):
            v_ptr = &v[i, 0]
            b_ptr = &b[i, 0]
            BDN_matvec_ptr(v_ptr, b_ptr, &ld[0], &dd[0], ud, N, strides)

cdef void CDDmat_matvec_ptr(T* v,
                            T* b,
                            real_t* ld,
                            real_t* ud,
                            int N,
                            int st):
    cdef:
        int i

    b[0] = ud[0]*v[st]
    b[(N-1)*st] = ld[N-2]*v[(N-2)*st]
    for i in xrange(1, N-1):
        b[i*st] = ud[i]*v[(i+1)*st] + ld[i-1]*v[(i-1)*st]

def CDD_matvec3D_ptr(T[:, :, ::1] v,
                     T[:, :, ::1] b,
                     real_t[::1] ld,
                     real_t[::1] ud,
                     int axis):
    cdef:
        int i, j, k, strides
        int N = ud.shape[0]+1
        T* v_ptr
        T* b_ptr

    strides = v.strides[axis]/v.itemsize
    if axis == 0:
        for j in range(v.shape[1]):
            for k in range(v.shape[2]):
                v_ptr = &v[0, j, k]
                b_ptr = &b[0, j, k]
                CDDmat_matvec_ptr(v_ptr, b_ptr, &ld[0], &ud[0], N, strides)
    elif axis == 1:
       for i in range(v.shape[0]):
            for k in range(v.shape[2]):
                v_ptr = &v[i, 0, k]
                b_ptr = &b[i, 0, k]
                CDDmat_matvec_ptr(v_ptr, b_ptr, &ld[0], &ud[0], N, strides)
    elif axis == 2:
        for i in range(v.shape[0]):
            for j in range(v.shape[1]):
                v_ptr = &v[i, j, 0]
                b_ptr = &b[i, j, 0]
                CDDmat_matvec_ptr(v_ptr, b_ptr, &ld[0], &ud[0], N, strides)

def CDD_matvec2D_ptr(T[:, ::1] v,
                     T[:, ::1] b,
                     real_t[::1] ld,
                     real_t[::1] ud,
                     int axis):
    cdef:
        int i, j, k, strides
        int N = ud.shape[0]+1
        T* v_ptr
        T* b_ptr

    strides = v.strides[axis]/v.itemsize
    if axis == 0:
        for j in range(v.shape[1]):
            v_ptr = &v[0, j]
            b_ptr = &b[0, j]
            CDDmat_matvec_ptr(v_ptr, b_ptr, &ld[0], &ud[0], N, strides)
    elif axis == 1:
       for i in range(v.shape[0]):
            v_ptr = &v[i, 0]
            b_ptr = &b[i, 0]
            CDDmat_matvec_ptr(v_ptr, b_ptr, &ld[0], &ud[0], N, strides)

def CDD_matvec1D_ptr(T[::1] v,
                     T[::1] b,
                     real_t[::1] ld,
                     real_t[::1] ud):
    cdef:
        int N = ud.shape[0]+1
        T* v_ptr = &v[0]
        T* b_ptr = &b[0]
    CDDmat_matvec_ptr(v_ptr, b_ptr, &ld[0], &ud[0], N, 1)

def SBB_matvec3D_ptr(T[:, :, ::1] v,
                     T[:, :, ::1] b,
                     real_t[::1] dd,
                     int axis):
    cdef:
        int i, j, k, strides
        int N = dd.shape[0]
        T* v_ptr
        T* b_ptr

    strides = v.strides[axis]/v.itemsize
    if axis == 0:
        for j in range(v.shape[1]):
            for k in range(v.shape[2]):
                v_ptr = &v[0, j, k]
                b_ptr = &b[0, j, k]
                SBB_matvec_ptr(v_ptr, b_ptr, &dd[0], N, strides)
    elif axis == 1:
       for i in range(v.shape[0]):
            for k in range(v.shape[2]):
                v_ptr = &v[i, 0, k]
                b_ptr = &b[i, 0, k]
                SBB_matvec_ptr(v_ptr, b_ptr, &dd[0], N, strides)
    elif axis == 2:
        for i in range(v.shape[0]):
            for j in range(v.shape[1]):
                v_ptr = &v[i, j, 0]
                b_ptr = &b[i, j, 0]
                SBB_matvec_ptr(v_ptr, b_ptr, &dd[0], N, strides)

def SBB_matvec2D_ptr(T[:, ::1] v,
                     T[:, ::1] b,
                     real_t[::1] dd,
                     int axis):
    cdef:
        int i, j, strides
        int N = dd.shape[0]
        T* v_ptr
        T* b_ptr

    strides = v.strides[axis]/v.itemsize
    if axis == 0:
        for j in range(v.shape[1]):
            v_ptr = &v[0, j]
            b_ptr = &b[0, j]
            SBB_matvec_ptr(v_ptr, b_ptr, &dd[0], N, strides)
    elif axis == 1:
       for i in range(v.shape[0]):
            v_ptr = &v[i, 0]
            b_ptr = &b[i, 0]
            SBB_matvec_ptr(v_ptr, b_ptr, &dd[0], N, strides)

cdef void SBB_matvec_ptr(T* v,
                         T* b,
                         real_t* dd,
                         int N,
                         int st):
    cdef:
        int i, j, k
        double p, r
        T d, s1, s2, o1, o2

    j = N-1
    s1 = 0.0
    s2 = 0.0
    o1 = 0.0
    o2 = 0.0
    b[j*st] = dd[j]*v[j*st]
    b[(j-1)*st] = dd[j-1]*v[(j-1)*st]
    for k in range(N-3, -1, -1):
        j = k+2
        p = k*dd[k]/(k+1)
        r = 24*(k+1)*(k+2)*M_PI
        d = v[j*st]/(j+3.)
        if k % 2 == 0:
            s1 += d
            s2 += (j+2)*(j+2)*d
            b[k*st] = dd[k]*v[k*st] + p*s1 + r*s2
        else:
            o1 += d
            o2 += (j+2)*(j+2)*d
            b[k*st] = dd[k]*v[k*st] + p*o1 + r*o2

def SBBmat_matvec(np.ndarray[T, ndim=1] v,
                  np.ndarray[T, ndim=1] b,
                  np.ndarray[real_t, ndim=1] dd):
    cdef:
        int i, j, k
        int N = v.shape[0]-4
        double p, r
        T d, s1, s2, o1, o2

    j = N-1
    s1 = 0.0
    s2 = 0.0
    o1 = 0.0
    o2 = 0.0
    b[j] = dd[j]*v[j]
    b[j-1] = dd[j-1]*v[j-1]
    for k in range(N-3, -1, -1):
        j = k+2
        p = k*dd[k]/(k+1)
        r = 24*(k+1)*(k+2)*np.pi
        d = v[j]/(j+3.)
        if k % 2 == 0:
            s1 += d
            s2 += (j+2)*(j+2)*d
            b[k] = dd[k]*v[k] + p*s1 + r*s2
        else:
            o1 += d
            o2 += (j+2)*(j+2)*d
            b[k] = dd[k]*v[k] + p*o1 + r*o2

def ADDmat_matvec(np.ndarray[T, ndim=1] v,
                  np.ndarray[T, ndim=1] b,
                  np.ndarray[real_t, ndim=1] dd):
    cdef:
        int i, j, k
        int N = v.shape[0]-2
        double p
        double pi = np.pi
        T d, s1, s2

    k = N-1
    s1 = 0.0
    s2 = 0.0
    b[k] = dd[k]*v[k]
    b[k-1] = dd[k-1]*v[k-1]
    for k in range(N-3, -1, -1):
        j = k+2
        p = -4*(k+1)*pi
        if j % 2 == 0:
            s1 += v[j]
            b[k] = dd[k]*v[k] + p*s1
        else:
            s2 += v[j]
            b[k] = dd[k]*v[k] + p*s2

def ADD_matvec(np.ndarray[T, ndim=1] v,
               np.ndarray[T, ndim=1] b,
               np.ndarray[real_t, ndim=1] dd):
    cdef:
        T* v_ptr=&v[0]
        T* b_ptr=&b[0]
    ADD_matvec_ptr(v_ptr, b_ptr, &dd[0], dd.shape[0], 1)

cdef void ADD_matvec_ptr(T* v,
                         T* b,
                         real_t* dd,
                         int N,
                         int st):
    cdef:
        int i, j, k
        double p
        T s1 = 0.0
        T s2 = 0.0
        T d

    k = N-1
    b[k*st] = dd[k]*v[k*st]
    b[(k-1)*st] = dd[k-1]*v[(k-1)*st]
    for k in range(N-3, -1, -1):
        j = k+2
        p = -4*(k+1)*M_PI
        if j % 2 == 0:
            s1 += v[j*st]
            b[k*st] = dd[k]*v[k*st] + p*s1
        else:
            s2 += v[j*st]
            b[k*st] = dd[k]*v[k*st] + p*s2

def ADD_matvec3D_ptr(T[:, :, ::1] v,
                     T[:, :, ::1] b,
                     real_t[::1] dd,
                     int axis):
    cdef:
        int i, j, k, strides
        int N = dd.shape[0]
        T* v_ptr
        T* b_ptr

    strides = v.strides[axis]/v.itemsize
    if axis == 0:
        for j in range(v.shape[1]):
            for k in range(v.shape[2]):
                v_ptr = &v[0, j, k]
                b_ptr = &b[0, j, k]
                ADD_matvec_ptr(v_ptr, b_ptr, &dd[0], N, strides)

    elif axis == 1:
       for i in range(v.shape[0]):
            for k in range(v.shape[2]):
                v_ptr = &v[i, 0, k]
                b_ptr = &b[i, 0, k]
                ADD_matvec_ptr(v_ptr, b_ptr, &dd[0], N, strides)
    elif axis == 2:
        for i in range(v.shape[0]):
            for j in range(v.shape[1]):
                v_ptr = &v[i, j, 0]
                b_ptr = &b[i, j, 0]
                ADD_matvec_ptr(v_ptr, b_ptr, &dd[0], N, strides)

def ADD_matvec2D_ptr(T[:, ::1] v,
                     T[:, ::1] b,
                     real_t[::1] dd,
                     int axis):
    cdef:
        int i, j, k, strides
        int N = dd.shape[0]
        T* v_ptr
        T* b_ptr

    strides = v.strides[axis]/v.itemsize
    if axis == 0:
        for j in range(v.shape[1]):
            v_ptr = &v[0, j]
            b_ptr = &b[0, j]
            ADD_matvec_ptr(v_ptr, b_ptr, &dd[0], N, strides)

    elif axis == 1:
       for i in range(v.shape[0]):
            v_ptr = &v[i, 0]
            b_ptr = &b[i, 0]
            ADD_matvec_ptr(v_ptr, b_ptr, &dd[0], N, strides)

cdef void ATT_matvec_ptr(T* v,
                         T* b,
                         int N,
                         int st):
    cdef:
        int i, j, k
        double p0, p1
        T s1 = 0.0
        T s2 = 0.0
        T s3 = 0.0
        T s4 = 0.0
        T d

    k = N-1
    b[(N-1)*st] = 0
    b[(N-2)*st] = 0
    for k in range(N-3, -1, -1):
        j = k+2
        p0 = M_PI/2
        p1 = M_PI/2*k**2
        if j % 2 == 0:
            s1 += j*v[j*st]
            s3 += j**3*v[j*st]
            b[k*st] = p0*s3 - p1*s1
        else:
            s2 += j*v[j*st]
            s4 += j**3*v[j*st]
            b[k*st] = p0*s4 - p1*s2

def ATT_matvec(np.ndarray[T, ndim=1] v,
               np.ndarray[T, ndim=1] b):
    cdef:
        T* v_ptr=&v[0]
        T* b_ptr=&b[0]
    ATT_matvec_ptr(v_ptr, b_ptr, v.shape[0], 1)

def ATT_matvec3D_ptr(T[:, :, ::1] v,
                     T[:, :, ::1] b,
                     int axis):
    cdef:
        int i, j, k, strides
        int N = v.shape[axis]
        T* v_ptr
        T* b_ptr

    strides = v.strides[axis]/v.itemsize
    if axis == 0:
        for j in range(v.shape[1]):
            for k in range(v.shape[2]):
                v_ptr = &v[0, j, k]
                b_ptr = &b[0, j, k]
                ATT_matvec_ptr(v_ptr, b_ptr, N, strides)

    elif axis == 1:
       for i in range(v.shape[0]):
            for k in range(v.shape[2]):
                v_ptr = &v[i, 0, k]
                b_ptr = &b[i, 0, k]
                ATT_matvec_ptr(v_ptr, b_ptr, N, strides)

    elif axis == 2:
        for i in range(v.shape[0]):
            for j in range(v.shape[1]):
                v_ptr = &v[i, j, 0]
                b_ptr = &b[i, j, 0]
                ATT_matvec_ptr(v_ptr, b_ptr, N, strides)

def ATT_matvec2D_ptr(T[:, ::1] v,
                     T[:, ::1] b,
                     int axis):
    cdef:
        int i, j, k, strides
        int N = v.shape[axis]
        T* v_ptr
        T* b_ptr

    strides = v.strides[axis]/v.itemsize
    if axis == 0:
        for j in range(v.shape[1]):
            v_ptr = &v[0, j]
            b_ptr = &b[0, j]
            ATT_matvec_ptr(v_ptr, b_ptr, N, strides)

    elif axis == 1:
       for i in range(v.shape[0]):
            v_ptr = &v[i, 0]
            b_ptr = &b[i, 0]
            ATT_matvec_ptr(v_ptr, b_ptr, N, strides)

cdef void GLL_matvec_ptr(T* v,
                         T* b,
                         int N,
                         int st):
    cdef:
        int i, j, k
        double p0, p1
        T s1 = 0.0
        T s2 = 0.0
        T s3 = 0.0
        T s4 = 0.0
        T d

    k = N-1
    b[(N-1)*st] = 0
    b[(N-2)*st] = 0
    for k in range(N-3, -1, -1):
        j = k+2
        p0 = 2*(k+0.5)/(2*k+1)
        p1 = p0*k*(k+1)
        if j % 2 == 0:
            s1 += j*(j+1)*v[j*st]
            s3 += v[j*st]
            b[k*st] = p0*s1 - p1*s3
        else:
            s2 += j*(j+1)*v[j*st]
            s4 += v[j*st]
            b[k*st] = p0*s2 - p1*s4

def GLL_matvec(np.ndarray[T, ndim=1] v,
               np.ndarray[T, ndim=1] b):
    cdef:
        T* v_ptr=&v[0]
        T* b_ptr=&b[0]
    GLL_matvec_ptr(v_ptr, b_ptr, v.shape[0], 1)

def GLL_matvec3D_ptr(T[:, :, ::1] v,
                     T[:, :, ::1] b,
                     int axis):
    cdef:
        int i, j, k, strides
        int N = v.shape[axis]
        T* v_ptr
        T* b_ptr

    strides = v.strides[axis]/v.itemsize
    if axis == 0:
        for j in range(v.shape[1]):
            for k in range(v.shape[2]):
                v_ptr = &v[0, j, k]
                b_ptr = &b[0, j, k]
                GLL_matvec_ptr(v_ptr, b_ptr, N, strides)

    elif axis == 1:
       for i in range(v.shape[0]):
            for k in range(v.shape[2]):
                v_ptr = &v[i, 0, k]
                b_ptr = &b[i, 0, k]
                GLL_matvec_ptr(v_ptr, b_ptr, N, strides)

    elif axis == 2:
        for i in range(v.shape[0]):
            for j in range(v.shape[1]):
                v_ptr = &v[i, j, 0]
                b_ptr = &b[i, j, 0]
                GLL_matvec_ptr(v_ptr, b_ptr, N, strides)

def GLL_matvec2D_ptr(T[:, ::1] v,
                     T[:, ::1] b,
                     int axis):
    cdef:
        int i, j, k, strides
        int N = v.shape[axis]
        T* v_ptr
        T* b_ptr

    strides = v.strides[axis]/v.itemsize
    if axis == 0:
        for j in range(v.shape[1]):
            v_ptr = &v[0, j]
            b_ptr = &b[0, j]
            GLL_matvec_ptr(v_ptr, b_ptr, N, strides)

    elif axis == 1:
       for i in range(v.shape[0]):
            v_ptr = &v[i, 0]
            b_ptr = &b[i, 0]
            GLL_matvec_ptr(v_ptr, b_ptr, N, strides)


cdef void CLL_matvec_ptr(T* v,
                         T* b,
                         int N,
                         int st):
    cdef:
        int i, j, k
        double p
        T s1 = 0.0
        T s2 = 0.0

    b[(N-1)*st] = 0
    for k in range(N-2, -1, -1):
        j = k+1
        if j % 2 == 0:
            s1 += v[j*st]
            b[k*st] = 2*s1
        else:
            s2 += v[j*st]
            b[k*st] = 2*s2

def CLL_matvec(np.ndarray[T, ndim=1] v,
               np.ndarray[T, ndim=1] b):
    cdef:
        T* v_ptr=&v[0]
        T* b_ptr=&b[0]
    CLL_matvec_ptr(v_ptr, b_ptr, v.shape[0], 1)

def CLL_matvec2D_ptr(T[:, ::1] v,
                     T[:, ::1] b,
                     int axis):
    cdef:
        int i, j, k, strides
        int N = v.shape[axis]
        T* v_ptr
        T* b_ptr

    strides = v.strides[axis]/v.itemsize
    if axis == 0:
        for j in range(v.shape[1]):
            v_ptr = &v[0, j]
            b_ptr = &b[0, j]
            CLL_matvec_ptr(v_ptr, b_ptr, N, strides)

    elif axis == 1:
       for i in range(v.shape[0]):
            v_ptr = &v[i, 0]
            b_ptr = &b[i, 0]
            CLL_matvec_ptr(v_ptr, b_ptr, N, strides)

def CLL_matvec3D_ptr(T[:, :, ::1] v,
                     T[:, :, ::1] b,
                     int axis):
    cdef:
        int i, j, k, strides
        int N = v.shape[axis]
        T* v_ptr
        T* b_ptr

    strides = v.strides[axis]/v.itemsize
    if axis == 0:
        for j in range(v.shape[1]):
            for k in range(v.shape[2]):
                v_ptr = &v[0, j, k]
                b_ptr = &b[0, j, k]
                CLL_matvec_ptr(v_ptr, b_ptr, N, strides)

    elif axis == 1:
       for i in range(v.shape[0]):
            for k in range(v.shape[2]):
                v_ptr = &v[i, 0, k]
                b_ptr = &b[i, 0, k]
                CLL_matvec_ptr(v_ptr, b_ptr, N, strides)

    elif axis == 2:
        for i in range(v.shape[0]):
            for j in range(v.shape[1]):
                v_ptr = &v[i, j, 0]
                b_ptr = &b[i, j, 0]
                CLL_matvec_ptr(v_ptr, b_ptr, N, strides)

cdef void CTT_matvec_ptr(T* v,
                         T* b,
                         int N,
                         int st):
    cdef:
        int i, j, k
        double p
        T s1 = 0.0
        T s2 = 0.0

    b[(N-1)*st] = 0
    for k in range(N-2, -1, -1):
        j = k+1
        if j % 2 == 0:
            s1 += (k+1)*v[j*st]
            b[k*st] = M_PI*s1
        else:
            s2 += (k+1)*v[j*st]
            b[k*st] = M_PI*s2

def CTT_matvec(np.ndarray[T, ndim=1] v,
               np.ndarray[T, ndim=1] b):
    cdef:
        T* v_ptr=&v[0]
        T* b_ptr=&b[0]
    CTT_matvec_ptr(v_ptr, b_ptr, v.shape[0], 1)

def CTT_matvec2D_ptr(T[:, ::1] v,
                     T[:, ::1] b,
                     int axis):
    cdef:
        int i, j, k, strides
        int N = v.shape[axis]
        T* v_ptr
        T* b_ptr

    strides = v.strides[axis]/v.itemsize
    if axis == 0:
        for j in range(v.shape[1]):
            v_ptr = &v[0, j]
            b_ptr = &b[0, j]
            CTT_matvec_ptr(v_ptr, b_ptr, N, strides)

    elif axis == 1:
       for i in range(v.shape[0]):
            v_ptr = &v[i, 0]
            b_ptr = &b[i, 0]
            CTT_matvec_ptr(v_ptr, b_ptr, N, strides)

def CTT_matvec3D_ptr(T[:, :, ::1] v,
                     T[:, :, ::1] b,
                     int axis):
    cdef:
        int i, j, k, strides
        int N = v.shape[axis]
        T* v_ptr
        T* b_ptr

    strides = v.strides[axis]/v.itemsize
    if axis == 0:
        for j in range(v.shape[1]):
            for k in range(v.shape[2]):
                v_ptr = &v[0, j, k]
                b_ptr = &b[0, j, k]
                CTT_matvec_ptr(v_ptr, b_ptr, N, strides)

    elif axis == 1:
       for i in range(v.shape[0]):
            for k in range(v.shape[2]):
                v_ptr = &v[i, 0, k]
                b_ptr = &b[i, 0, k]
                CTT_matvec_ptr(v_ptr, b_ptr, N, strides)

    elif axis == 2:
        for i in range(v.shape[0]):
            for j in range(v.shape[1]):
                v_ptr = &v[i, j, 0]
                b_ptr = &b[i, j, 0]
                CTT_matvec_ptr(v_ptr, b_ptr, N, strides)

cdef void Tridiagonal_matvec_ptr(T* v,
                                 T* b,
                                 real_t* ld,
                                 real_t* dd,
                                 real_t* ud,
                                 int N,
                                 int st):
    cdef:
        int i

    b[0] = dd[0]*v[0] + ud[0]*v[2*st]
    b[st] = dd[1]*v[st] + ud[1]*v[3*st]
    for i in xrange(2, N-2):
        b[i*st] = ld[i-2]*v[(i-2)*st] + dd[i]*v[i*st] + ud[i]*v[(i+2)*st]
    i = N-2
    b[i*st] = ld[i-2]*v[(i-2)*st] + dd[i]*v[i*st]
    i = N-1
    b[i*st] = ld[i-2]* v[(i-2)*st] + dd[i]*v[i*st]

def Tridiagonal_matvec2D_ptr(T[:, ::1] v,
                             T[:, ::1] b,
                             real_t[::1] ld,
                             real_t[::1] dd,
                             real_t[::1] ud,
                             int axis):
    cdef:
        int i, j, strides
        int N = dd.shape[0]
        T* v_ptr
        T* b_ptr

    strides = v.strides[axis]/v.itemsize
    if axis == 0:
        for j in range(v.shape[1]):
            v_ptr = &v[0, j]
            b_ptr = &b[0, j]
            Tridiagonal_matvec_ptr(v_ptr, b_ptr, &ld[0], &dd[0], &ud[0], N, strides)
    elif axis == 1:
        for i in range(v.shape[0]):
            v_ptr = &v[i, 0]
            b_ptr = &b[i, 0]
            Tridiagonal_matvec_ptr(v_ptr, b_ptr, &ld[0], &dd[0], &ud[0], N, strides)

def Tridiagonal_matvec3D_ptr(T[:, :, ::1] v,
                             T[:, :, ::1] b,
                             real_t[::1] ld,
                             real_t[::1] dd,
                             real_t[::1] ud,
                             int axis):
    cdef:
        int i, j, k, strides
        int N = dd.shape[0]
        T* v_ptr
        T* b_ptr

    strides = v.strides[axis]/v.itemsize
    if axis == 0:
        for j in range(v.shape[1]):
            for k in range(v.shape[2]):
                v_ptr = &v[0, j, k]
                b_ptr = &b[0, j, k]
                Tridiagonal_matvec_ptr(v_ptr, b_ptr, &ld[0], &dd[0], &ud[0], N, strides)
    elif axis == 1:
        for i in range(v.shape[0]):
            for k in range(v.shape[2]):
                v_ptr = &v[i, 0, k]
                b_ptr = &b[i, 0, k]
                Tridiagonal_matvec_ptr(v_ptr, b_ptr, &ld[0], &dd[0], &ud[0], N, strides)
    elif axis == 2:
        for i in range(v.shape[0]):
            for j in range(v.shape[1]):
                v_ptr = &v[i, j, 0]
                b_ptr = &b[i, j, 0]
                Tridiagonal_matvec_ptr(v_ptr, b_ptr, &ld[0], &dd[0], &ud[0], N, strides)

def Tridiagonal_matvec(T[::1] v,
                       T[::1] b,
                       real_t[::1] ld,
                       real_t[::1] dd,
                       real_t[::1] ud):
    cdef:
        np.intp_t i
        np.intp_t N = dd.shape[0]

    b[0] = dd[0]*v[0] + ud[0]*v[2]
    b[1] = dd[1]*v[1] + ud[1]*v[3]
    for i in xrange(2, N-2):
        b[i] = ld[i-2]* v[i-2] + dd[i]*v[i] + ud[i]*v[i+2]
    i = N-2
    b[i] = ld[i-2]* v[i-2] + dd[i]*v[i]
    i = N-1
    b[i] = ld[i-2]* v[i-2] + dd[i]*v[i]

def Pentadiagonal_matvec(T[::1] v,
                         T[::1] b,
                         real_t[::1] ldd,
                         real_t[::1] ld,
                         real_t[::1] dd,
                         real_t[::1] ud,
                         real_t[::1] udd):
    cdef:
        int i
        int N = dd.shape[0]

    b[0] = dd[0]*v[0] + ud[0]*v[2] + udd[0]*v[4]
    b[1] = dd[1]*v[1] + ud[1]*v[3] + udd[1]*v[5]
    b[2] = ld[0]*v[0] + dd[2]*v[2] + ud[2]*v[4] + udd[2]*v[6]
    b[3] = ld[1]*v[1] + dd[3]*v[3] + ud[3]*v[5] + udd[3]*v[7]
    for i in xrange(4, N-4):
        b[i] = ldd[i-4]*v[i-4] + ld[i-2]*v[i-2] + dd[i]*v[i] + ud[i]*v[i+2] + udd[i]*v[i+4]
    i = N-4
    b[i] = ldd[i-4]*v[i-4] + ld[i-2]*v[i-2] + dd[i]*v[i] + ud[i]*v[i+2]
    i = N-3
    b[i] = ldd[i-4]*v[i-4] + ld[i-2]*v[i-2] + dd[i]*v[i] + ud[i]*v[i+2]
    i = N-2
    b[i] = ldd[i-4]*v[i-4] + ld[i-2]*v[i-2] + dd[i]*v[i]
    i = N-1
    b[i] = ldd[i-4]*v[i-4] + ld[i-2]*v[i-2] + dd[i]*v[i]

cdef void Pentadiagonal_matvec_ptr(T* v,
                                   T* b,
                                   real_t* ldd,
                                   real_t* ld,
                                   real_t* dd,
                                   real_t* ud,
                                   real_t* udd,
                                   int N,
                                   int st):
    cdef:
        int i

    b[0] = dd[0]*v[0] + ud[0]*v[2*st] + udd[0]*v[4*st]
    b[1*st] = dd[1]*v[1*st] + ud[1]*v[3*st] + udd[1]*v[5*st]
    b[2*st] = ld[0]*v[0] + dd[2]*v[2*st] + ud[2]*v[4*st] + udd[2]*v[6*st]
    b[3*st] = ld[1]*v[1*st] + dd[3]*v[3*st] + ud[3]*v[5*st] + udd[3]*v[7*st]
    for i in xrange(4, N-4):
        b[i*st] = ldd[i-4]*v[(i-4)*st] + ld[i-2]*v[(i-2)*st] + dd[i]*v[i*st] + ud[i]*v[(i+2)*st] + udd[i]*v[(i+4)*st]
    i = N-4
    b[i*st] = ldd[i-4]*v[(i-4)*st] + ld[i-2]*v[(i-2)*st] + dd[i]*v[i*st] + ud[i]*v[(i+2)*st]
    i = N-3
    b[i*st] = ldd[i-4]*v[(i-4)*st] + ld[i-2]*v[(i-2)*st] + dd[i]*v[i*st] + ud[i]*v[(i+2)*st]
    i = N-2
    b[i*st] = ldd[i-4]*v[(i-4)*st] + ld[i-2]*v[(i-2)*st] + dd[i]*v[i*st]
    i = N-1
    b[i*st] = ldd[i-4]*v[(i-4)*st] + ld[i-2]*v[(i-2)*st] + dd[i]*v[i*st]

def Pentadiagonal_matvec3D_ptr(T[:, :, ::1] v,
                               T[:, :, ::1] b,
                               real_t[::1] ldd,
                               real_t[::1] ld,
                               real_t[::1] dd,
                               real_t[::1] ud,
                               real_t[::1] udd,
                               np.int64_t axis):
    cdef:
        int i, j, k, strides
        int N = dd.shape[0]
        T* v_ptr
        T* b_ptr

    strides = v.strides[axis]/v.itemsize
    if axis == 0:
        for j in range(v.shape[1]):
            for k in range(v.shape[2]):
                v_ptr = &v[0, j, k]
                b_ptr = &b[0, j, k]
                Pentadiagonal_matvec_ptr(v_ptr, b_ptr, &ldd[0], &ld[0], &dd[0],
                                         &ud[0], &udd[0], N, strides)

    elif axis == 1:
       for i in range(v.shape[0]):
            for k in range(v.shape[2]):
                v_ptr = &v[i, 0, k]
                b_ptr = &b[i, 0, k]
                Pentadiagonal_matvec_ptr(v_ptr, b_ptr, &ldd[0], &ld[0], &dd[0],
                                         &ud[0], &udd[0], N, strides)

    elif axis == 2:
        for i in range(v.shape[0]):
            for j in range(v.shape[1]):
                v_ptr = &v[i, j, 0]
                b_ptr = &b[i, j, 0]
                Pentadiagonal_matvec_ptr(v_ptr, b_ptr, &ldd[0], &ld[0], &dd[0],
                                         &ud[0], &udd[0], N, strides)

def Pentadiagonal_matvec2D_ptr(T[:, ::1] v,
                               T[:, ::1] b,
                               real_t[::1] ldd,
                               real_t[::1] ld,
                               real_t[::1] dd,
                               real_t[::1] ud,
                               real_t[::1] udd,
                               int axis):
    cdef:
        int i, j, strides
        int N = dd.shape[0]
        T* v_ptr
        T* b_ptr

    strides = v.strides[axis]/v.itemsize
    if axis == 0:
        for j in range(v.shape[1]):
            v_ptr = &v[0, j]
            b_ptr = &b[0, j]
            Pentadiagonal_matvec_ptr(v_ptr, b_ptr, &ldd[0], &ld[0], &dd[0],
                                     &ud[0], &udd[0], N, strides)

    elif axis == 1:
       for i in range(v.shape[0]):
            v_ptr = &v[i, 0]
            b_ptr = &b[i, 0]
            Pentadiagonal_matvec_ptr(v_ptr, b_ptr, &ldd[0], &ld[0], &dd[0],
                                     &ud[0], &udd[0], N, strides)

def CBD_matvec(np.ndarray[T, ndim=1] v,
               np.ndarray[T, ndim=1] b,
               np.ndarray[real_t, ndim=1] ld,
               np.ndarray[real_t, ndim=1] ud,
               np.ndarray[real_t, ndim=1] udd):
    cdef:
        int i
        int N = udd.shape[0]

    b[0] = ud[0]*v[1] + udd[0]*v[3]
    for i in xrange(1, N):
        b[i] = ld[i-1]* v[i-1] + ud[i]*v[i+1] + udd[i]*v[i+3]
    i = N
    b[i] = ld[i-1]* v[i-1] + ud[i]*v[i+1]

cdef void CBD_matvec_ptr(T* v,
                         T* b,
                         real_t* ld,
                         real_t* ud,
                         real_t* udd,
                         int N,
                         int st):
    cdef:
        int i

    b[0] = ud[0]*v[1*st] + udd[0]*v[3*st]
    for i in xrange(1, N):
        b[i*st] = ld[i-1]*v[(i-1)*st] + ud[i]*v[(i+1)*st] + udd[i]*v[(i+3)*st]
    i = N
    b[i*st] = ld[i-1]*v[(i-1)*st] + ud[i]*v[(i+1)*st]

def CBD_matvec2D_ptr(T[:, ::1] v,
                     T[:, ::1] b,
                     real_t [::1] ld,
                     real_t [::1] ud,
                     real_t [::1] udd,
                     int axis):
    cdef:
        int i, j, strides
        int N = udd.shape[0]
        T* v_ptr
        T* b_ptr

    strides = v.strides[axis]/v.itemsize
    if axis == 0:
        for j in range(v.shape[1]):
            v_ptr = &v[0, j]
            b_ptr = &b[0, j]
            CBD_matvec_ptr(v_ptr, b_ptr, &ld[0], &ud[0], &udd[0], N, strides)
    elif axis == 1:
        for i in range(v.shape[0]):
            v_ptr = &v[i, 0]
            b_ptr = &b[i, 0]
            CBD_matvec_ptr(v_ptr, b_ptr, &ld[0], &ud[0], &udd[0], N, strides)

def CBD_matvec3D_ptr(T[:, :, ::1] v,
                     T[:, :, ::1] b,
                     real_t [::1] ld,
                     real_t [::1] ud,
                     real_t [::1] udd,
                     int axis):
    cdef:
        int i, j, k, strides
        int N = udd.shape[0]
        T* v_ptr
        T* b_ptr

    strides = v.strides[axis]/v.itemsize
    if axis == 0:
        for j in range(v.shape[1]):
            for k in range(v.shape[2]):
                v_ptr = &v[0, j, k]
                b_ptr = &b[0, j, k]
                CBD_matvec_ptr(v_ptr, b_ptr, &ld[0], &ud[0], &udd[0], N, strides)
    elif axis == 1:
        for i in range(v.shape[0]):
            for k in range(v.shape[2]):
                v_ptr = &v[i, 0, k]
                b_ptr = &b[i, 0, k]
                CBD_matvec_ptr(v_ptr, b_ptr, &ld[0], &ud[0], &udd[0], N, strides)
    elif axis == 2:
        for i in range(v.shape[0]):
            for j in range(v.shape[1]):
                v_ptr = &v[i, j, 0]
                b_ptr = &b[i, j, 0]
                CBD_matvec_ptr(v_ptr, b_ptr, &ld[0], &ud[0], &udd[0], N, strides)

cdef void CDB_matvec_ptr(T* v,
                         T* b,
                         real_t* lld,
                         real_t* ld,
                         real_t* ud,
                         int N,
                         int st):
    cdef:
        int i, j, k

    b[0] = ud[0]*v[st]
    for k in xrange(1, 3):
        b[k*st] = ld[k-1]*v[(k-1)*st] + ud[k]*v[(k+1)*st]
    for k in xrange(3, N):
        b[k*st] = lld[k-3]*v[(k-3)*st] + ld[k-1]*v[(k-1)*st] + ud[k]*v[(k+1)*st]
    for k in xrange(N, N+2):
        b[k*st] = lld[k-3]*v[(k-3)*st] + ld[k-1]* v[(k-1)*st]
    b[(N+2)*st] = lld[N-1]*v[(N-1)*st]

def CDB_matvec(T[::1] v,
               T[::1] b,
               real_t [::1] lld,
               real_t [::1] ld,
               real_t [::1] ud):
    cdef:
        int N = ud.shape[0]
        T* v_ptr = &v[0]
        T* b_ptr = &b[0]
    CDB_matvec_ptr(v_ptr, b_ptr, &lld[0], &ld[0], &ud[0], N, 1)

def CDB_matvec2D_ptr(T[:, ::1] v,
                     T[:, ::1] b,
                     real_t[::1] lld,
                     real_t[::1] ld,
                     real_t[::1] ud,
                     int axis):
    cdef:
        int i, j, strides
        int N = ud.shape[0]
        T* v_ptr
        T* b_ptr

    strides = v.strides[axis]/v.itemsize
    if axis == 0:
        for j in range(v.shape[1]):
            v_ptr = &v[0, j]
            b_ptr = &b[0, j]
            CDB_matvec_ptr(v_ptr, b_ptr, &lld[0], &ld[0], &ud[0], N, strides)
    elif axis == 1:
        for i in range(v.shape[0]):
            v_ptr = &v[i, 0]
            b_ptr = &b[i, 0]
            CDB_matvec_ptr(v_ptr, b_ptr, &lld[0], &ld[0], &ud[0], N, strides)

def CDB_matvec3D_ptr(T[:, :, ::1] v,
                     T[:, :, ::1] b,
                     real_t[::1] lld,
                     real_t[::1] ld,
                     real_t[::1] ud,
                     int axis):
    cdef:
        int i, j, k, strides
        int N = ud.shape[0]
        T* v_ptr
        T* b_ptr

    strides = v.strides[axis]/v.itemsize
    if axis == 0:
        for j in range(v.shape[1]):
            for k in range(v.shape[2]):
                v_ptr = &v[0, j, k]
                b_ptr = &b[0, j, k]
                CDB_matvec_ptr(v_ptr, b_ptr, &lld[0], &ld[0], &ud[0], N, strides)
    elif axis == 1:
        for i in range(v.shape[0]):
            for k in range(v.shape[2]):
                v_ptr = &v[i, 0, k]
                b_ptr = &b[i, 0, k]
                CDB_matvec_ptr(v_ptr, b_ptr, &lld[0], &ld[0], &ud[0], N, strides)
    elif axis == 2:
        for i in range(v.shape[0]):
            for j in range(v.shape[1]):
                v_ptr = &v[i, j, 0]
                b_ptr = &b[i, j, 0]
                CDB_matvec_ptr(v_ptr, b_ptr, &lld[0], &ld[0], &ud[0], N, strides)

cdef void BBD_matvec_ptr(T* v,
                         T* b,
                         real_t ld,
                         real_t* dd,
                         real_t* ud,
                         real_t* uud,
                         int N,
                         int st):
    cdef:
        int i

    b[0] = dd[0]*v[0] + ud[0]*v[2*st] + uud[0]*v[4*st]
    b[st] = dd[1]*v[st] + ud[1]*v[3*st] + uud[1]*v[5*st]

    for k in range(2, N):
        b[k*st] = ld*v[(k-2)*st] + dd[k]*v[k*st] + ud[k]*v[(k+2)*st] + uud[k]*v[(k+4)*st]

    for k in range(N, N+2):
        b[k*st] = ld*v[(k-2)*st] + dd[k]*v[k*st] + ud[k]*v[(k+2)*st]

def BBD_matvec3D_ptr(T[:, :, ::1] v,
                     T[:, :, ::1] b,
                     real_t ld,
                     real_t[::1] dd,
                     real_t[::1] ud,
                     real_t[::1] uud,
                     int axis):
    cdef:
        int i, j, k, strides
        int N = uud.shape[0]
        T* v_ptr
        T* b_ptr

    strides = v.strides[axis]/v.itemsize
    if axis == 0:
        for j in range(v.shape[1]):
            for k in range(v.shape[2]):
                v_ptr = &v[0, j, k]
                b_ptr = &b[0, j, k]
                BBD_matvec_ptr(v_ptr, b_ptr, ld, &dd[0], &ud[0], &uud[0], N, strides)
    elif axis == 1:
        for i in range(v.shape[0]):
            for k in range(v.shape[2]):
                v_ptr = &v[i, 0, k]
                b_ptr = &b[i, 0, k]
                BBD_matvec_ptr(v_ptr, b_ptr, ld, &dd[0], &ud[0], &uud[0], N, strides)
    elif axis == 2:
        for i in range(v.shape[0]):
            for j in range(v.shape[1]):
                v_ptr = &v[i, j, 0]
                b_ptr = &b[i, j, 0]
                BBD_matvec_ptr(v_ptr, b_ptr, ld, &dd[0], &ud[0], &uud[0], N, strides)

def BBD_matvec2D_ptr(T[:, ::1] v,
                     T[:, ::1] b,
                     real_t ld,
                     real_t[::1] dd,
                     real_t[::1] ud,
                     real_t[::1] uud,
                     int axis):
    cdef:
        int i, j, strides
        int N = uud.shape[0]
        T* v_ptr
        T* b_ptr

    strides = v.strides[axis]/v.itemsize
    if axis == 0:
        for j in range(v.shape[1]):
            v_ptr = &v[0, j]
            b_ptr = &b[0, j]
            BBD_matvec_ptr(v_ptr, b_ptr, ld, &dd[0], &ud[0], &uud[0], N, strides)
    elif axis == 1:
        for i in range(v.shape[0]):
            v_ptr = &v[i, 0]
            b_ptr = &b[i, 0]
            BBD_matvec_ptr(v_ptr, b_ptr, ld, &dd[0], &ud[0], &uud[0], N, strides)

def BBD_matvec1D_ptr(T[::1] v,
                     T[::1] b,
                     real_t ld,
                     real_t[::1] dd,
                     real_t[::1] ud,
                     real_t[::1] uud):
    cdef:
        int i, j, strides
        int N = uud.shape[0]
        T* v_ptr = &v[0]
        T* b_ptr = &b[0]
    BBD_matvec_ptr(v_ptr, b_ptr, ld, &dd[0], &ud[0], &uud[0], N, 1)

def Helmholtz_matvec_1D(np.ndarray[T, ndim=1] v,
                        np.ndarray[T, ndim=1] b,
                        real_t alfa,
                        real_t beta,
                        np.ndarray[real_t, ndim=1] dd,
                        np.ndarray[real_t, ndim=1] ud,
                        np.ndarray[real_t, ndim=1] bd):
    # b = (alfa*A + beta*B)*v
    # For B matrix ld = ud = -pi/2
    cdef:
        int i, j, k
        int N = dd.shape[0]
        T s1 = 0.0
        T s2 = 0.0
        double p

    k = N-1
    b[k] = (dd[k]*alfa + bd[k]*beta)*v[k] - M_PI_2*beta*v[k-2]
    b[k-1] = (dd[k-1]*alfa + bd[k-1]*beta)*v[k-1] - M_PI_2*beta*v[k-3]

    for k in range(N-3, 1, -1):
        p = ud[k]*alfa
        if k % 2 == 0:
            s2 += v[k+2]
            b[k] = (dd[k]*alfa + bd[k]*beta)*v[k] - M_PI_2*beta*(v[k-2] + v[k+2]) + p*s2
        else:
            s1 += v[k+2]
            b[k] = (dd[k]*alfa + bd[k]*beta)*v[k] - M_PI_2*beta*(v[k-2] + v[k+2]) + p*s1

    k = 1
    s1 += v[k+2]
    s2 += v[k+1]
    b[k] = (dd[k]*alfa + bd[k]*beta)*v[k] - M_PI_2*beta*v[k+2] + ud[k]*alfa*s1
    b[k-1] = (dd[k-1]*alfa + bd[k-1]*beta)*v[k-1] - M_PI_2*beta*v[k+1] + ud[k-1]*alfa*s2

cdef void Helmholtz_matvec_ptr(T* v,
                               T* b,
                               real_t alfa,
                               real_t beta,
                               real_t* dd,
                               real_t* ud,
                               real_t* bd,
                               int N,
                               int st):
    # b = (alfa*A + beta*B)*v
    # For B matrix ld = ud = -pi/2
    cdef:
        int i, j, k
        T s1 = 0.0
        T s2 = 0.0
        double p

    k = N-1
    b[k*st] = (dd[k]*alfa + bd[k]*beta)*v[k*st] - M_PI_2*beta*v[(k-2)*st]
    b[(k-1)*st] = (dd[k-1]*alfa + bd[k-1]*beta)*v[(k-1)*st] - M_PI_2*beta*v[(k-3)*st]

    for k in range(N-3, 1, -1):
        p = ud[k]*alfa
        if k % 2 == 0:
            s2 += v[(k+2)*st]
            b[k*st] = (dd[k]*alfa + bd[k]*beta)*v[k*st] - M_PI_2*beta*(v[(k-2)*st] + v[(k+2)*st]) + p*s2
        else:
            s1 += v[(k+2)*st]
            b[k*st] = (dd[k]*alfa + bd[k]*beta)*v[k*st] - M_PI_2*beta*(v[(k-2)*st] + v[(k+2)*st]) + p*s1

    k = 1
    s1 += v[(k+2)*st]
    s2 += v[(k+1)*st]
    b[k*st] = (dd[k]*alfa + bd[k]*beta)*v[k*st] - M_PI_2*beta*v[(k+2)*st] + ud[k]*alfa*s1
    b[(k-1)*st] = (dd[k-1]*alfa + bd[k-1]*beta)*v[(k-1)*st] - M_PI_2*beta*v[(k+1)*st] + ud[k-1]*alfa*s2

def Helmholtz_matvec3D_ptr(T[:, :, ::1] v,
                           T[:, :, ::1] b,
                           real_t[:, :, ::1] alfa,
                           real_t[:, :, ::1] beta,
                           # 3 upper diagonals of SBB
                           real_t[::1] dd,
                           real_t[::1] ud,
                           real_t[::1] bd,
                           int axis):
    cdef:
        int i, j, k, strides
        int N = dd.shape[0]
        T* v_ptr
        T* b_ptr

    strides = v.strides[axis]/v.itemsize
    if axis == 0:
        for j in range(v.shape[1]):
            for k in range(v.shape[2]):
                v_ptr = &v[0, j, k]
                b_ptr = &b[0, j, k]
                Helmholtz_matvec_ptr(v_ptr, b_ptr, alfa[0, j, k],
                                     beta[0, j, k], &dd[0], &ud[0], &bd[0], N, strides)

    elif axis == 1:
       for i in range(v.shape[0]):
            for k in range(v.shape[2]):
                v_ptr = &v[i, 0, k]
                b_ptr = &b[i, 0, k]
                Helmholtz_matvec_ptr(v_ptr, b_ptr, alfa[i, 0, k],
                                     beta[i, 0, k], &dd[0], &ud[0], &bd[0], N, strides)

    elif axis == 2:
        for i in range(v.shape[0]):
            for j in range(v.shape[1]):
                v_ptr = &v[i, j, 0]
                b_ptr = &b[i, j, 0]
                Helmholtz_matvec_ptr(v_ptr, b_ptr, alfa[i, j, 0],
                                     beta[i, j, 0], &dd[0], &ud[0], &bd[0], N, strides)

def Helmholtz_matvec2D_ptr(T[:, ::1] v,
                           T[:, ::1] b,
                           real_t[:, ::1] alfa,
                           real_t[:, ::1] beta,
                           # 3 upper diagonals of SBB
                           real_t[::1] dd,
                           real_t[::1] ud,
                           real_t[::1] bd,
                           int axis):
    cdef:
        int i, j, strides
        int N = dd.shape[0]
        T* v_ptr
        T* b_ptr

    strides = v.strides[axis]/v.itemsize
    if axis == 0:
        for j in range(v.shape[1]):
            v_ptr = &v[0, j]
            b_ptr = &b[0, j]
            Helmholtz_matvec_ptr(v_ptr, b_ptr, alfa[0, j],
                                 beta[0, j], &dd[0], &ud[0], &bd[0], N, strides)

    elif axis == 1:
       for i in range(v.shape[0]):
            v_ptr = &v[i, 0]
            b_ptr = &b[i, 0]
            Helmholtz_matvec_ptr(v_ptr, b_ptr, alfa[i, 0],
                                 beta[i, 0], &dd[0], &ud[0], &bd[0], N, strides)

def Helmholtz_matvec(v, b, alfa, beta, A, B, axis):
    dd = A[0].copy()
    ud = A[2].copy()
    bd = B[0].copy()
    if v.ndim == 1:
        Helmholtz_matvec_1D(v, b, alfa, beta, dd, ud, bd)
    elif v.ndim == 2:
        Helmholtz_matvec2D_ptr(v, b, alfa, beta, dd, ud, bd, axis)
    elif v.ndim == 3:
        Helmholtz_matvec3D_ptr(v, b, alfa, beta, dd, ud, bd, axis)

def Helmholtz_Neumann_matvec_1D(np.ndarray[T, ndim=1] v,
                        np.ndarray[T, ndim=1] b,
                        real_t alfa,
                        real_t beta,
                        np.ndarray[real_t, ndim=1] dd,
                        np.ndarray[real_t, ndim=1] ud,
                        np.ndarray[real_t, ndim=1] bl,
                        np.ndarray[real_t, ndim=1] bd,
                        np.ndarray[real_t, ndim=1] bu):
    # b = (alfa*A + beta*B)*v
    # A matrix has diagonal dd and upper second diagonal at ud
    # B matrix has diagonal bd and second upper and lower diagonals bu and bl
    cdef:
        int i, j, k, j2
        int N = dd.shape[0]
        T s1 = 0.0
        T s2 = 0.0
        double p

    for k in (N-1, N-2):
        b[k] = (dd[k]*alfa + bd[k]*beta)*v[k]*k**2 + bl[k-2]*beta*v[k-2]*(k-2)**2

    for k in range(N-3, 1, -1):
        p = ud[k]*alfa
        if k % 2 == 0:
            s2 += v[k+2]*(k+2)**2
            b[k] = (dd[k]*alfa + bd[k]*beta)*v[k]*k**2 + beta*(bl[k-2]*v[k-2]*(k-2)**2 + bu[k]*v[k+2]*(k+2)**2) + p*s2
        else:
            s1 += v[k+2]*(k+2)**2
            b[k] = (dd[k]*alfa + bd[k]*beta)*v[k]*k**2 + beta*(bl[k-2]*v[k-2]*(k-2)**2 + bu[k]*v[k+2]*(k+2)**2) + p*s1

    k = 1
    s1 += v[k+2]*(k+2)**2
    b[k] = (dd[k]*alfa + bd[k]*beta)*v[k]*k**2 + beta*(bu[k]*v[k+2]*(k+2)**2) + ud[k]*alfa*s1
    k = 0
    s2 += v[k+2]*(k+2)**2
    b[k] = (dd[k]*alfa + bd[k]*beta)*v[k]*k**2 + beta*(bu[k]*v[k+2]*(k+2)**2) + ud[k]*alfa*s2
    b[0] += bd[0]*v[0]*beta
    b[2] += bl[0]*v[0]*beta


cdef void Helmholtz_Neumann_matvec_ptr(T* v,
                        T* b,
                        real_t alfa,
                        real_t beta,
                        real_t* dd,
                        real_t* ud,
                        real_t* bl,
                        real_t* bd,
                        real_t* bu,
                        int N,
                        int st):
    # b = (alfa*A + beta*B)*v
    # A matrix has diagonal dd and upper second diagonal at ud
    # B matrix has diagonal bd and second upper and lower diagonals bu and bl
    cdef:
        int i, j, k, j2
        T s1 = 0.0
        T s2 = 0.0
        double p

    for k in (N-1, N-2):
        j2 = k*k
        b[k*st] = (dd[k]*alfa + bd[k]*beta)*v[k*st]*j2 + bl[k-2]*beta*v[(k-2)*st]*j2

    for k in range(N-3, 1, -1):
        p = ud[k]*alfa
        if k % 2 == 0:
            s2 += v[(k+2)*st]*(k+2)**2
            b[k*st] = (dd[k]*alfa + bd[k]*beta)*v[k*st]*k**2 + beta*(bl[k-2]*v[(k-2)*st]*(k-2)**2 + bu[k]*v[(k+2)*st]*(k+2)**2) + p*s2
        else:
            s1 += v[(k+2)*st]*(k+2)**2
            b[k*st] = (dd[k]*alfa + bd[k]*beta)*v[k*st]*k**2 + beta*(bl[k-2]*v[(k-2)*st]*(k-2)**2 + bu[k]*v[(k+2)*st]*(k+2)**2) + p*s1

    k = 1
    s1 += v[(k+2)*st]*(k+2)**2
    b[k*st] = (dd[k]*alfa + bd[k]*beta)*v[k*st]*k**2 + beta*(bu[k]*v[(k+2)*st]*(k+2)**2) + ud[k]*alfa*s1
    k = 0
    s2 += v[(k+2)*st]*(k+2)**2
    b[k*st] = (dd[k]*alfa + bd[k]*beta)*v[k*st]*k**2 + beta*(bu[k]*v[(k+2)*st]*(k+2)**2) + ud[k]*alfa*s2
    b[0] += bd[0]*v[0]*beta
    b[2*st] += bl[0]*v[0]*beta

def Helmholtz_Neumann_matvec3D_ptr(T[:, :, ::1] v,
                           T[:, :, ::1] b,
                           real_t[:, :, ::1] alfa,
                           real_t[:, :, ::1] beta,
                           # 3 upper diagonals of SBB
                           real_t[::1] dd,
                           real_t[::1] ud,
                           real_t[::1] bl,
                           real_t[::1] bd,
                           real_t[::1] bu,
                           int axis):
    cdef:
        int i, j, k, strides
        int N = dd.shape[0]
        T* v_ptr
        T* b_ptr

    strides = v.strides[axis]/v.itemsize
    if axis == 0:
        for j in range(v.shape[1]):
            for k in range(v.shape[2]):
                v_ptr = &v[0, j, k]
                b_ptr = &b[0, j, k]
                Helmholtz_Neumann_matvec_ptr(v_ptr, b_ptr, alfa[0, j, k],
                                     beta[0, j, k], &dd[0], &ud[0], &bl[0],
                                     &bd[0], &bu[0], N, strides)

    elif axis == 1:
       for i in range(v.shape[0]):
            for k in range(v.shape[2]):
                v_ptr = &v[i, 0, k]
                b_ptr = &b[i, 0, k]
                Helmholtz_Neumann_matvec_ptr(v_ptr, b_ptr, alfa[i, 0, k],
                                     beta[i, 0, k], &dd[0], &ud[0], &bl[0],
                                     &bd[0], &bu[0], N, strides)

    elif axis == 2:
        for i in range(v.shape[0]):
            for j in range(v.shape[1]):
                v_ptr = &v[i, j, 0]
                b_ptr = &b[i, j, 0]
                Helmholtz_Neumann_matvec_ptr(v_ptr, b_ptr, alfa[i, j, 0],
                                     beta[i, j, 0], &dd[0], &ud[0], &bl[0],
                                     &bd[0], &bu[0], N, strides)


def Helmholtz_Neumann_matvec2D_ptr(T[:, ::1] v,
                           T[:, ::1] b,
                           real_t[:, ::1] alfa,
                           real_t[:, ::1] beta,
                           real_t[::1] dd,
                           real_t[::1] ud,
                           real_t[::1] bl,
                           real_t[::1] bd,
                           real_t[::1] bu,
                           int axis):
    cdef:
        int i, j, strides
        int N = dd.shape[0]
        T* v_ptr
        T* b_ptr

    strides = v.strides[axis]/v.itemsize
    if axis == 0:
        for j in range(v.shape[1]):
            v_ptr = &v[0, j]
            b_ptr = &b[0, j]
            Helmholtz_Neumann_matvec_ptr(v_ptr, b_ptr, alfa[0, j],
                                 beta[0, j], &dd[0], &ud[0], &bl[0], &bd[0], &bu[0], N, strides)

    elif axis == 1:
       for i in range(v.shape[0]):
            v_ptr = &v[i, 0]
            b_ptr = &b[i, 0]
            Helmholtz_Neumann_matvec_ptr(v_ptr, b_ptr, alfa[i, 0],
                                 beta[i, 0], &dd[0], &ud[0], &bl[0], &bd[0], &bu[0], N, strides)

def Helmholtz_Neumann_matvec1D_ptr(T[::1] v,
                           T[::1] b,
                           real_t alfa,
                           real_t beta,
                           real_t[::1] dd,
                           real_t[::1] ud,
                           real_t[::1] bl,
                           real_t[::1] bd,
                           real_t[::1] bu):
    cdef:
        int i, j, strides
        int N = dd.shape[0]
        T* v_ptr
        T* b_ptr
    v_ptr = &v[0]
    b_ptr = &b[0]
    strides = v.strides[0]/v.itemsize
    Helmholtz_Neumann_matvec_ptr(v_ptr, b_ptr, alfa,
                                 beta, &dd[0], &ud[0],
                                 &bl[0], &bd[0], &bu[0], N, strides)


def Helmholtz_Neumann_matvec(v, b, alfa, beta, A, B, axis):
    N = A[0].shape[0]
    k = np.arange(N)
    j2 = k**2
    j2[0] = 1
    j2 = 1/j2
    j2[0] = 0
    A_0 = A[0]*j2
    A_2 = A[2]*j2[2:]
    j2[0] = 1
    B_0 = B[0]*j2
    B_2 = B[2]*j2[2:]
    B_m2 = B[-2]*j2[:-2]

    if v.ndim == 1:
        Helmholtz_Neumann_matvec_1D(v, b, alfa, beta, A_0, A_2, B_m2, B_0, B_2)
        #Helmholtz_Neumann_matvec1D_ptr(v, b, alfa, beta, A_0, A_2, B_m2, B_0, B_2)
    elif v.ndim == 2:
        Helmholtz_Neumann_matvec2D_ptr(v, b, alfa, beta, A_0, A_2, B_m2, B_0, B_2, axis)
    elif v.ndim == 3:
        Helmholtz_Neumann_matvec3D_ptr(v, b, alfa, beta, A_0, A_2, B_m2, B_0, B_2, axis)


cdef void Biharmonic_matvec_ptr(T* v,
                                T* b,
                                real_t a0,
                                real_t alfa,
                                real_t beta,
                                # 3 upper diagonals of SBB
                                real_t* sii,
                                real_t* siu,
                                real_t* siuu,
                                # All 3 diagonals of ABB
                                real_t* ail,
                                real_t* aii,
                                real_t* aiu,
                                # All 5 diagonals of BBB
                                real_t* bill,
                                real_t* bil,
                                real_t* bii,
                                real_t* biu,
                                real_t* biuu,
                                int N,
                                int st):
    cdef:
        int i, j, k
        vector[double] ldd, ld, dd, ud, udd
        double p, r
        T d, s1, s2, o1, o2

    dd.resize(N)
    ld.resize(N)
    ldd.resize(N)
    ud.resize(N)
    udd.resize(N)

    for i in xrange(N):
        dd[i] = a0*sii[i] + alfa*aii[i] + beta*bii[i]

    for i in xrange(N-2):
        ld[i] = alfa*ail[i] + beta*bil[i]

    for i in xrange(N-4):
        ldd[i] = beta*bill[i]

    for i in xrange(N-2):
        ud[i] = a0*siu[i] + alfa*aiu[i] + beta*biu[i]

    for i in xrange(N-4):
        udd[i] = a0*siuu[i] + beta*biuu[i]

    i = N-1
    b[i*st] = ldd[i-4]*v[(i-4)*st]+ ld[i-2]* v[(i-2)*st] + dd[i]*v[i*st]
    i = N-2
    b[i*st] = ldd[i-4]*v[(i-4)*st]+ ld[i-2]* v[(i-2)*st] + dd[i]*v[i*st]
    i = N-3
    b[i*st] = ldd[i-4]*v[(i-4)*st]+ ld[i-2]* v[(i-2)*st] + dd[i]*v[i*st] + ud[i]*v[(i+2)*st]
    i = N-4
    b[i*st] = ldd[i-4]*v[(i-4)*st]+ ld[i-2]* v[(i-2)*st] + dd[i]*v[i*st] + ud[i]*v[(i+2)*st]
    i = N-5
    b[i*st] = ldd[i-4]*v[(i-4)*st]+ ld[i-2]* v[(i-2)*st] + dd[i]*v[i*st] + ud[i]*v[(i+2)*st] + udd[i]*v[(i+4)*st]
    i = N-6
    b[i*st] = ldd[i-4]*v[(i-4)*st]+ ld[i-2]* v[(i-2)*st] + dd[i]*v[i*st] + ud[i]*v[(i+2)*st] + udd[i]*v[(i+4)*st]

    s1 = 0.0
    s2 = 0.0
    o1 = 0.0
    o2 = 0.0
    for k in xrange(N-7, -1, -1):
        j = k+6
        p = k*sii[k]/(k+1.)
        r = 24*(k+1)*(k+2)*M_PI
        d = v[j*st]/(j+3.)
        if k % 2 == 0:
            s1 += d
            s2 += (j+2)*(j+2)*d
            b[k*st] = (p*s1 + r*s2)*a0
        else:
            o1 += d
            o2 += (j+2)*(j+2)*d
            b[k*st] = (p*o1 + r*o2)*a0

        if k > 3:
            b[k*st] += ldd[k-4]*v[(k-4)*st]+ ld[k-2]* v[(k-2)*st] + dd[k]*v[k*st] + ud[k]*v[(k+2)*st] + udd[k]*v[(k+4)*st]
        elif k > 1:
            b[k*st] += ld[k-2]* v[(k-2)*st] + dd[k]*v[k*st] + ud[k]*v[(k+2)*st] + udd[k]*v[(k+4)*st]
        else:
            b[k*st] += dd[k]*v[k*st] + ud[k]*v[(k+2)*st] + udd[k]*v[(k+4)*st]

def Biharmonic_matvec3D_ptr(T[:, :, ::1] v,
                            T[:, :, ::1] b,
                            real_t a0,
                            real_t[:, :, ::1] alfa,
                            real_t[:, :, ::1] beta,
                            # 3 upper diagonals of SBB
                            real_t[::1] sii,
                            real_t[::1] siu,
                            real_t[::1] siuu,
                            # All 3 diagonals of ABB
                            real_t[::1] ail,
                            real_t[::1] aii,
                            real_t[::1] aiu,
                            # All 5 diagonals of BBB
                            real_t[::1] bill,
                            real_t[::1] bil,
                            real_t[::1] bii,
                            real_t[::1] biu,
                            real_t[::1] biuu,
                            int axis):
    cdef:
        int i, j, k, strides
        int N = sii.shape[0]
        T* v_ptr
        T* b_ptr

    strides = v.strides[axis]/v.itemsize
    if axis == 0:
        for j in range(v.shape[1]):
            for k in range(v.shape[2]):
                v_ptr = &v[0, j, k]
                b_ptr = &b[0, j, k]
                Biharmonic_matvec_ptr(v_ptr, b_ptr, a0, alfa[0, j, k],
                                      beta[0, j, k], &sii[0], &siu[0], &siuu[0], &ail[0], &aii[0],
                                      &aiu[0], &bill[0], &bil[0], &bii[0], &biu[0], &biuu[0], N, strides)

    elif axis == 1:
       for i in range(v.shape[0]):
            for k in range(v.shape[2]):
                v_ptr = &v[i, 0, k]
                b_ptr = &b[i, 0, k]
                Biharmonic_matvec_ptr(v_ptr, b_ptr, a0, alfa[i, 0, k],
                                      beta[i, 0, k], &sii[0], &siu[0], &siuu[0], &ail[0], &aii[0],
                                      &aiu[0], &bill[0], &bil[0], &bii[0], &biu[0], &biuu[0], N, strides)

    elif axis == 2:
        for i in range(v.shape[0]):
            for j in range(v.shape[1]):
                v_ptr = &v[i, j, 0]
                b_ptr = &b[i, j, 0]
                Biharmonic_matvec_ptr(v_ptr, b_ptr, a0, alfa[i, j, 0],
                                      beta[i, j, 0], &sii[0], &siu[0], &siuu[0], &ail[0], &aii[0],
                                      &aiu[0], &bill[0], &bil[0], &bii[0], &biu[0], &biuu[0], N, strides)

def Biharmonic_matvec2D_ptr(T[:, ::1] v,
                            T[:, ::1] b,
                            real_t a0,
                            real_t[:, ::1] alfa,
                            real_t[:, ::1] beta,
                            # 3 upper diagonals of SBB
                            real_t[::1] sii,
                            real_t[::1] siu,
                            real_t[::1] siuu,
                            # All 3 diagonals of ABB
                            real_t[::1] ail,
                            real_t[::1] aii,
                            real_t[::1] aiu,
                            # All 5 diagonals of BBB
                            real_t[::1] bill,
                            real_t[::1] bil,
                            real_t[::1] bii,
                            real_t[::1] biu,
                            real_t[::1] biuu,
                            int axis):
    cdef:
        int i, j, k, strides
        int N = sii.shape[0]
        T* v_ptr
        T* b_ptr

    strides = v.strides[axis]/v.itemsize
    if axis == 0:
        for j in range(v.shape[1]):
            v_ptr = &v[0, j]
            b_ptr = &b[0, j]
            Biharmonic_matvec_ptr(v_ptr, b_ptr, a0, alfa[0, j],
                                  beta[0, j], &sii[0], &siu[0], &siuu[0], &ail[0], &aii[0],
                                  &aiu[0], &bill[0], &bil[0], &bii[0], &biu[0], &biuu[0], N, strides)

    elif axis == 1:
       for i in range(v.shape[0]):
            v_ptr = &v[i, 0]
            b_ptr = &b[i, 0]
            Biharmonic_matvec_ptr(v_ptr, b_ptr, a0, alfa[i, 0],
                                  beta[i, 0], &sii[0], &siu[0], &siuu[0], &ail[0], &aii[0],
                                  &aiu[0], &bill[0], &bil[0], &bii[0], &biu[0], &biuu[0], N, strides)

def Biharmonic_matvec_1D(T[::1] v,
                         T[::1] b,
                         real_t a0,
                         real_t[::1] alfa,
                         real_t[::1] beta,
                         # 3 upper diagonals of SBB
                         real_t[::1] sii,
                         real_t[::1] siu,
                         real_t[::1] siuu,
                         # All 3 diagonals of ABB
                         real_t[::1] ail,
                         real_t[::1] aii,
                         real_t[::1] aiu,
                         # All 5 diagonals of BBB
                         real_t[::1] bill,
                         real_t[::1] bil,
                         real_t[::1] bii,
                         real_t[::1] biu,
                         real_t[::1] biuu,
                         int axis):
    cdef:
        int i, j, k, strides
        int N = sii.shape[0]
        T* v_ptr
        T* b_ptr

    strides = v.strides[axis]/v.itemsize
    v_ptr = &v[0]
    b_ptr = &b[0]
    Biharmonic_matvec_ptr(v_ptr, b_ptr, a0, alfa[0],
                          beta[0], &sii[0], &siu[0], &siuu[0], &ail[0], &aii[0],
                          &aiu[0], &bill[0], &bil[0], &bii[0], &biu[0], &biuu[0], N, strides)

def Biharmonic_matvec(v, b, a0, alfa, beta, sii, siu, siuu, ail, aii,
                      aiu, bill, bil, bii, biu, biuu, axis=0):
    if v.ndim == 1:
        Biharmonic_matvec_1D(v, b, a0, alfa, beta, sii, siu, siuu, ail, aii,
                             aiu, bill, bil, bii, biu, biuu)
    elif v.ndim == 2:
        Biharmonic_matvec2D_ptr(v, b, a0, alfa, beta, sii, siu, siuu, ail, aii,
                                aiu, bill, bil, bii, biu, biuu, axis)
    elif v.ndim == 3:
        Biharmonic_matvec3D_ptr(v, b, a0, alfa, beta, sii, siu, siuu, ail, aii,
                                aiu, bill, bil, bii, biu, biuu, axis)



r"""
Module for some integrators.

    - IRK3:     Implicit third order Runge-Kutta
    - RK4:      Runge-Kutta fourth order
    - ETD:      Exponential time differencing Euler method
    - ETDRK4:   Exponential time differencing Runge-Kutta fourth order

See, e.g.,
H. Montanelli and N. Bootland "Solving periodic semilinear PDEs in 1D, 2D and
3D with exponential integrators", https://arxiv.org/pdf/1604.08900.pdf

Integrators are set up to solve equations like

.. math::

    \frac{\partial u}{\partial t} = L u + N(u)

where :math:`u` is the solution, :math:`L` is a linear operator and
:math:`N(u)` is the nonlinear part of the right hand side.

Note
----
`RK4`, `ETD` and `ETDRK4` can only be used with Fourier function spaces,
as they assume all matrices are diagonal.

"""
import types
import numpy as np
from shenfun import Function, TPMatrix, TrialFunction, TestFunction, inner, la

__all__ = ('IRK3', 'RK4', 'ETDRK4', 'ETD')

#pylint: disable=unused-variable

class IntegratorBase:
    """Abstract base class for integrators

    Parameters
    ----------
        T : TensorProductSpace
        L : function
            To compute linear part of right hand side
        N : function
            To compute nonlinear part of right hand side
        update : function
                 To be called at the end of a timestep
        params : dictionary
                 Any relevant keyword arguments

    """

    def __init__(self, T,
                 L=None,
                 N=None,
                 update=None,
                 **params):
        _p = {'call_update': -1,
              'dt': 0}
        _p.update(params)
        self.params = _p
        self.T = T
        if L is not None:
            self.LinearRHS = types.MethodType(L, self)
        if N is not None:
            self.NonlinearRHS = types.MethodType(N, self)
        if update is not None:
            self.update = types.MethodType(update, self)

    def update(self, u, u_hat, t, tstep, **par):
        pass

    def LinearRHS(self, *args, **kwargs):
        pass

    def NonlinearRHS(self, *args, **kwargs):
        pass

    def setup(self, dt):
        """Set up solver"""
        pass

    def solve(self, u, u_hat, dt, trange):
        """Integrate forward in end_time

        Parameters
        ----------
            u : array
                The solution array in physical space
            u_hat : array
                    The solution array in spectral space
            dt : float
                 Timestep
            trange : two-tuple
                     Time and end time
        """
        pass


class IRK3(IntegratorBase):
    """Third order implicit Runge Kutta

    Parameters
    ----------
        T : TensorProductSpace
        L : function of TrialFunction(T)
            To compute linear part of right hand side
        N : function
            To compute nonlinear part of right hand side
        update : function
                 To be called at the end of a timestep
        params : dictionary
                 Any relevant keyword arguments

    """
    def __init__(self, T,
                 L=None,
                 N=None,
                 update=None,
                 **params):
        IntegratorBase.__init__(self, T, L=L, N=N, update=update, **params)
        self.dU = Function(T)
        self.dU1 = Function(T)
        self.a = (8./15., 5./12., 3./4.)
        self.b = (0.0, -17./60., -5./12.)
        self.c = (0., 8./15., 2./3., 1)
        self.solver = None
        self.rhs_mats = None
        self.w0 = Function(self.T).v
        self.mask = self.T.get_mask_nyquist()

    def setup(self, dt):
        self.params['dt'] = dt
        u = TrialFunction(self.T)
        v = TestFunction(self.T)

        # Note that we are here assembling implicit left hand side matrices,
        # as well as matrices that can be used to assemble the right hande side
        # much faster through matrix-vector products

        a, b = self.a, self.b
        self.solver = []
        for rk in range(3):
            mats = inner(v, u - ((a[rk]+b[rk])*dt/2)*self.LinearRHS(u))
            if len(mats[0].naxes) == 1:
                self.solver.append(la.SolverGeneric1ND(mats))
            elif len(mats[0].naxes) == 2:
                self.solver.append(la.SolverGeneric2ND(mats))
            else:
                raise NotImplementedError

        self.rhs_mats = []
        for rk in range(3):
            self.rhs_mats.append(inner(v, u + ((a[rk]+b[rk])*dt/2)*self.LinearRHS(u)))

        self.mass = inner(u, v)

    def compute_rhs(self, u, u_hat, dU, dU1, rk):
        a = self.a[rk]
        b = self.b[rk]
        dt = self.params['dt']
        dU = self.NonlinearRHS(u, u_hat, dU, **self.params)
        dU.mask_nyquist(self.mask)
        w1 = dU*a*dt + self.dU1*b*dt
        self.dU1[:] = dU
        return w1

    def solve(self, u, u_hat, dt, trange):
        if self.solver is None or abs(self.params['dt']-dt) > 1e-12:
            self.setup(dt)
        t, end_time = trange
        tstep = 0
        while t < end_time-1e-8:
            for rk in range(3):
                dU = self.compute_rhs(u, u_hat, self.dU, self.dU1, rk)
                for mat in self.rhs_mats[rk]:
                    w0 = mat.matvec(u_hat, self.w0)
                    dU += w0
                u_hat = self.solver[rk](dU, u_hat)
                u_hat.mask_nyquist(self.mask)

            t += self.dt
            tstep += 1
            self.update(u, u_hat, t, tstep, **self.params)


class ETD(IntegratorBase):
    """Exponential time differencing Euler method

    H. Montanelli and N. Bootland "Solving periodic semilinear PDEs in 1D, 2D and
    3D with exponential integrators", https://arxiv.org/pdf/1604.08900.pdf

    Parameters
    ----------
        T : TensorProductSpace
        L : function
            To compute linear part of right hand side
        N : function
            To compute nonlinear part of right hand side
        update : function
            To be called at the end of a timestep
        params : dictionary
            Any relevant keyword arguments
    """

    def __init__(self, T,
                 L=None,
                 N=None,
                 update=None,
                 **params):
        IntegratorBase.__init__(self, T, L=L, N=N, update=update, **params)
        self.dU = Function(T)
        self.psi = None
        self.ehL = None

    def setup(self, dt):
        """Set up ETD ODE solver"""
        self.params['dt'] = dt
        L = self.LinearRHS(**self.params)
        if isinstance(L, TPMatrix):
            assert L.isidentity()
            L = L.scale
        L = np.atleast_1d(L)
        hL = L*dt
        self.ehL = np.exp(hL)
        M = 50
        psi = self.psi = np.zeros(hL.shape, dtype=float)
        for k in range(1, M+1):
            ll = hL+np.exp(np.pi*1j*(k-0.5)/M)
            psi += ((np.exp(ll)-1.)/ll).real

        psi /= M

    def solve(self, u, u_hat, dt, trange):
        """Integrate forward in time

        Parameters
        ----------
            u : array
                The solution array in physical space
            u_hat : array
                The solution array in spectral space
            dt : float
                Timestep
            trange : two-tuple
                Time and end time
        """
        if self.psi is None or abs(self.params['dt']-dt) > 1e-12:
            self.setup(dt)
        t, end_time = trange
        tstep = 0
        while t < end_time-1e-8:
            t += dt
            tstep += 1
            self.dU = self.NonlinearRHS(u, u_hat, self.dU, **self.params)
            u_hat[:] = self.ehL*u_hat + dt*self.psi*self.dU
            self.update(u, u_hat, t, tstep, **self.params)
        return u_hat


class ETDRK4(IntegratorBase):
    """Exponential time differencing Runge-Kutta 4'th order method

    H. Montanelli and N. Bootland "Solving periodic semilinear PDEs in 1D, 2D and
    3D with exponential integrators", https://arxiv.org/pdf/1604.08900.pdf

    Parameters
    ----------
        T : TensorProductSpace
        L : function
            To compute linear part of right hand side
        N : function
            To compute nonlinear part of right hand side
        update : function
            To be called at the end of a timestep
        params : dictionary
            Any relevant keyword arguments
    """
    def __init__(self, T,
                 L=None,
                 N=None,
                 update=None,
                 **params):
        IntegratorBase.__init__(self, T, L=L, N=N, update=update, **params)
        self.U_hat0 = Function(T)
        self.U_hat1 = Function(T)
        self.dU = Function(T)
        self.dU0 = Function(T)
        self.V2 = Function(T)
        self.psi = np.zeros((4,)+self.U_hat0.shape, dtype=float)
        self.a = None
        self.b = [0.5, 0.5, 0.5]
        self.ehL = None
        self.ehL_h = None

    def setup(self, dt):
        """Set up ETDRK4 ODE solver"""
        self.params['dt'] = dt
        L = self.LinearRHS(**self.params)
        if isinstance(L, TPMatrix):
            assert L.isidentity()
            L = L.scale
        L = np.atleast_1d(L)
        hL = L*dt
        self.ehL = np.exp(hL)
        self.ehL_h = np.exp(hL/2.)
        M = 50
        psi = self.psi = np.zeros((4,) + hL.shape, dtype=float)
        for k in range(1, M+1):
            ll = hL+np.exp(np.pi*1j*(k-0.5)/M)
            psi[0] += ((np.exp(ll)-1.)/ll).real
            psi[1] += ((np.exp(ll)-ll-1.)/ll**2).real
            psi[2] += ((np.exp(ll)-0.5*ll**2-ll-1.)/ll**3).real
            ll2 = hL/2.+np.exp(np.pi*1j*(k-0.5)/M)
            psi[3] += ((np.exp(ll2)-1.)/(ll2)).real

        psi /= M
        a = [psi[0]-3*psi[1]+4*psi[2]]
        a.append(2*psi[1]-4*psi[2])
        a.append(2*psi[1]-4*psi[2])
        a.append(-psi[1]+4*psi[2])
        self.a = a

    def solve(self, u, u_hat, dt, trange):
        """Integrate forward in time

        Parameters
        ----------
            u : array
                The solution array in physical space
            u_hat : array
                The solution array in spectral space
            dt : float
                Timestep
            trange : two-tuple
                Time and end time
        """
        if self.a is None or abs(self.params['dt']-dt) > 1e-12:
            self.setup(dt)
        t, end_time = trange
        tstep = 0
        while t < end_time-1e-8:
            t += dt
            tstep += 1

            self.U_hat0[:] = u_hat*self.ehL_h
            self.U_hat1[:] = u_hat*self.ehL
            for rk in range(4):
                self.dU = self.NonlinearRHS(u, u_hat, self.dU, **self.params)
                if rk < 2:
                    u_hat[:] = self.U_hat0 + self.b[rk]*dt*self.psi[3]*self.dU
                elif rk == 2:
                    u_hat[:] = self.ehL_h*self.V2 + self.b[rk]*dt*self.psi[3]*(2*self.dU-self.dU0)

                if rk == 0:
                    self.dU0[:] = self.dU
                    self.V2[:] = u_hat

                self.U_hat1 += self.a[rk]*dt*self.dU
            u_hat[:] = self.U_hat1
            self.update(u, u_hat, t, tstep, **self.params)
        return u_hat


class RK4(IntegratorBase):
    """Regular 4'th order Runge-Kutta integrator

    Parameters
    ----------
        T : TensorProductSpace
        L : function
            To compute linear part of right hand side
        N : function
            To compute nonlinear part of right hand side
        update : function
            To be called at the end of a timestep
        params : dictionary
            Any relevant keyword arguments
    """
    def __init__(self, T,
                 L=None,
                 N=None,
                 update=None,
                 **params):
        IntegratorBase.__init__(self, T, L=L, N=N, update=update, **params)
        self.U_hat0 = Function(T)
        self.U_hat1 = Function(T)
        self.dU = Function(T)
        self.a = np.array([1./6., 1./3., 1./3., 1./6.])
        self.b = np.array([0.5, 0.5, 1.])

    def setup(self, dt):
        """Set up RK4 ODE solver"""
        self.params['dt'] = dt

    def solve(self, u, u_hat, dt, trange):
        """Integrate forward in end_time

        Parameters
        ----------
            u : array
                The solution array in physical space
            u_hat : array
                The solution array in spectral space
            dt : float
                Timestep
            trange : two-tuple
                Time and end time
        """
        if self.a is None or abs(self.params['dt']-dt) > 1e-12:
            self.setup(dt)
        t, end_time = trange
        tstep = 0
        L = self.LinearRHS(**self.params)
        while t < end_time-1e-8:
            t += dt
            tstep += 1
            self.U_hat0[:] = self.U_hat1[:] = u_hat
            for rk in range(4):
                dU = self.NonlinearRHS(u, u_hat, self.dU, **self.params)
                if L:
                    dU += L*u_hat
                if rk < 3:
                    u_hat[:] = self.U_hat0 + self.b[rk]*dt*dU
                self.U_hat1 += self.a[rk]*dt*dU
            u_hat[:] = self. U_hat1
            self.update(u, u_hat, t, tstep, **self.params)
        return u_hat



"""
Quasi-inverse matrices for Tau and Galerkin methods

@article{julien09,
title = {Efficient multi-dimensional solution of PDEs using Chebyshev spectral methods},
journal = {Journal of Computational Physics},
volume = {228},
number = {5},
pages = {1480-1503},
year = {2009},
issn = {0021-9991},
doi = {https://doi.org/10.1016/j.jcp.2008.10.043}
}
"""
import numpy as np
from shenfun.matrixbase import SparseMatrix

__all__ = ('QIGmat', 'QITmat', 'QICGmat', 'QICTmat')

# Note to self. Matrices below do not use ck because we use them in
# scalar products not inverted with (T_j, T_k)_w = ck*pi/2 \delta_{kj}

class QImat(SparseMatrix):

    def __mul__(self, y):
        """Returns copy of self.__mul__(y) <==> self*y"""
        if isinstance(y, SparseMatrix):
            return y.__quasi__(self)
        return SparseMatrix.__mul__(self, y)

class QIGmat(QImat):
    """Quasi-inverse matrix for the Galerkin method

    Parameters
    ----------
    N : int
        The number of quadrature points

    """
    def __init__(self, N):
        k = np.arange(N)
        #d = {
        #    0: 1/4/(k[2:]*(k[2:]-1)),
        #    2: -1/2/(k[2:-2]**2-1),
        #    4: 1/4/(k[2:-4]*(k[2:-4]+1))}
        #SparseMatrix.__init__(self, d, (N-2, N-2))
        d = {
            0: 1/4/(k[2:]*(k[2:]-1)),
            2: -1/2/(k[2:]**2-1),
            4: 1/4/(k[2:-2]*(k[2:-2]+1))}
        # Note: truncating the upper diagonals is in agreement with
        # \cite{julien09}.
        d[2][-2:] = 0
        d[4][-2:] = 0
        SparseMatrix.__init__(self, d, (N-2, N))

class QITmat(QImat):
    """Quasi-inverse matrix for the Tau method

    Parameters
    ----------
    N : int
        The number of quadrature points

    """
    def __init__(self, N):
        k = np.arange(N)
        d = {
            -2: 1/4/(k[2:]*(k[2:]-1)),
            0: np.zeros(N),
            2: np.zeros(N-2)}
        d[0][2:-2] = -1/2/(k[2:-2]**2-1)
        d[2][2:-2] = 1/4/(k[2:-4]*(k[2:-4]+1))

        # Truncate, like \cite{julien09}
        d[0][-2:] = 0
        d[2][-2:] = 0
        SparseMatrix.__init__(self, d, (N, N))

class QICGmat(QImat):
    """Quasi-inverse matrix for the Galerkin method

    Parameters
    ----------
    N : int
        The number of quadrature points

    """
    def __init__(self, N):
        k = np.arange(N)
        d = {
            0: 1/(2*(k[:-1]+1)),
            2: -1/(2*(k[:-2]+1))}
        SparseMatrix.__init__(self, d, (N-1, N))

class QICTmat(QImat):
    """Quasi-inverse matrix for the Tau method.

    Parameters
    ----------
    N : int
        The number of quadrature points

    """
    def __init__(self, N):
        k = np.arange(N)
        kk = np.arange(N)
        kk[0] = 1
        dk = np.ones(N)
        dk[0] = 0
        d = {
            -1: 1/(2*(k[:-1]+1)),
             1: -dk[:-1]/(2*(kk[:-1]))}
        SparseMatrix.__init__(self, d, (N, N))
